"""Configuration for the GW tests"""

from typing import Callable

from nessai.livepoint import dict_to_live_points
from nessai.model import Model as BaseModel
import pytest


@pytest.fixture()
def injection_parameters():
    injection_parameters = dict(
        mass_ratio=0.9,
        chirp_mass=25.0,
        a_1=0.4,
        a_2=0.3,
        tilt_1=0.5,
        tilt_2=1.0,
        phi_12=1.7,
        phi_jl=0.3,
        luminosity_distance=2000.0,
        theta_jn=0.4,
        psi=2.659,
        phase=1.3,
        geocent_time=1126259642.413,
        ra=1.375,
        dec=-1.2108,
    )
    return injection_parameters


@pytest.fixture()
def get_bilby_gw_model() -> Callable:
    """Return a function will provide a nessai model given parameters
    and an injection.
    """

    def get_model(parameters, injection_parameters) -> BaseModel:

        import bilby

        priors = bilby.gw.prior.BBHPriorDict()
        fixed_params = [
            "chirp_mass",
            "mass_ratio",
            "phi_12",
            "phi_jl",
            "a_1",
            "a_2",
            "tilt_1",
            "tilt_2",
            "ra",
            "dec",
            "luminosity_distance",
            "geocent_time",
            "theta_jn",
            "psi",
            "phase",
        ]
        try:
            fixed_params.remove(parameters)
        except ValueError:
            for p in parameters:
                fixed_params.remove(p)
        priors["geocent_time"] = bilby.core.prior.Uniform(
            minimum=injection_parameters["geocent_time"] - 0.1,
            maximum=injection_parameters["geocent_time"] + 0.1,
            name="geocent_time",
            latex_label="$t_c$",
            unit="$s$",
        )
        for key in fixed_params:
            if key in injection_parameters:
                priors[key] = injection_parameters[key]

        waveform_generator = bilby.gw.WaveformGenerator(
            duration=1,
            sampling_frequency=256,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,  # noqa
        )

        likelihood = bilby.gw.GravitationalWaveTransient(
            interferometers=["H1"],
            waveform_generator=waveform_generator,
            priors=priors,
            phase_marginalization="phase" not in fixed_params,
            distance_marginalization=False,
            time_marginalization=False,
            reference_frame="sky",
        )

        likelihood = bilby.core.likelihood.ZeroLikelihood(likelihood)

        def log_prior(theta):
            return priors.ln_prob(theta, axis=0)

        search_parameter_keys = []
        for key in priors:
            if (
                isinstance(priors[key], bilby.core.prior.Prior)
                and priors[key].is_fixed is False
            ):
                search_parameter_keys.append(key)

        def log_likelihood(theta):
            params = {key: t for key, t in zip(search_parameter_keys, theta)}

            likelihood.parameters.update(params)
            return likelihood.log_likelihood_ratio()

        class Model(BaseModel):
            def __init__(self, names, priors):
                self.names = names
                self.priors = priors
                self._update_bounds()

            @staticmethod
            def log_likelihood(x, **kwargs):
                theta = [x[n].item() for n in search_parameter_keys]
                return log_likelihood(theta)

            @staticmethod
            def log_prior(x, names=None, **kwargs):
                if names is None:
                    names = search_parameter_keys
                theta = {n: x[n] for n in names}
                return log_prior(theta)

            def _update_bounds(self):
                self.bounds = {
                    key: [self.priors[key].minimum, self.priors[key].maximum]
                    for key in self.names
                }

            def new_point(self, N=1):
                prior_samples = self.priors.sample(size=N)
                samples = {n: prior_samples[n] for n in self.names}
                self._update_bounds()
                return dict_to_live_points(samples)

            def new_point_log_prob(self, x):
                return self.log_prior(x)

        return Model(search_parameter_keys, priors)

    return get_model



"""Tests for level-related methods"""

import os
from unittest.mock import MagicMock

from nessai.samplers.importancesampler import (
    ImportanceNestedSampler as INS,
    OrderedSamples,
)
import numpy as np
import pytest


@pytest.mark.parametrize("include_likelihood", [False, True])
@pytest.mark.parametrize("use_log_weights", [False, True])
def test_determine_threshold_entropy(
    ins, samples, include_likelihood, use_log_weights
):
    samples = np.sort(samples, order="logL")
    ins.plot = False
    n = INS.determine_threshold_entropy(
        ins,
        samples,
        q=0.5,
        use_log_weights=use_log_weights,
        include_likelihood=include_likelihood,
    )
    assert 0 < n < samples.size


def test_determine_threshold_entropy_plot(ins, samples, tmp_path):
    samples = np.sort(samples, order="logL")
    ins.plot = True
    ins._plot_level_cdf = True
    ins.output = tmp_path / "test_entropy_plot"
    ins.plot_level_cdf = MagicMock()
    ins.iteration = 2
    n = INS.determine_threshold_entropy(
        ins,
        samples,
        q=0.5,
    )
    assert 0 < n < samples.size
    ins.plot_level_cdf.assert_called_once()
    assert os.path.exists(os.path.join(ins.output, "levels", "level_2"))


@pytest.mark.parametrize("include_likelihood", [False, True])
def test_determine_threshold_quantile(ins, samples, include_likelihood):
    samples = np.sort(samples, order="logL")
    ins.plot = False
    n = INS.determine_threshold_quantile(
        ins,
        samples,
        q=0.8,
        include_likelihood=include_likelihood,
    )
    assert 0 < n < samples.size


@pytest.mark.parametrize(
    "n, min_remove, min_samples, n_live, expected",
    [
        [10, 5, 8, 20, 10],
        [4, 5, 8, 20, 5],
        [15, 5, 10, 20, 10],
    ],
)
@pytest.mark.parametrize("method", ["entropy", "quantile"])
def test_determine_threshold(
    ins,
    n,
    min_remove,
    min_samples,
    n_live,
    expected,
    method,
):
    ins.min_samples = min_samples
    ins.min_remove = min_remove
    ins.max_samples = None
    ins.draw_constant = True
    ins.nlive = n_live
    samples = np.empty(n_live, dtype=[("x", "f8"), ("logL", "f8")])
    samples["logL"] = 10 * np.arange(n_live)

    expected_log_l = samples["logL"][expected]

    ins.determine_threshold_quantile = MagicMock(return_value=n)
    ins.determine_threshold_entropy = MagicMock(return_value=n)

    out = INS.determine_log_likelihood_threshold(
        ins, samples, method=method, q=0.8
    )

    if method == "entropy":
        ins.determine_threshold_entropy.assert_called_once_with(samples, q=0.8)
        ins.determine_threshold_quantile.assert_not_called()
    else:
        ins.determine_threshold_quantile.assert_called_once_with(
            samples, q=0.8
        )
        ins.determine_threshold_entropy.assert_not_called()

    assert out == expected_log_l


@pytest.mark.parametrize(
    "n_samples, n_remove, min_remove, "
    "min_samples, max_samples, n_live, expected",
    [
        [50, 10, 5, 10, 55, 30, 25],
        [56, 10, 5, 10, 55, 30, 31],
        [50, 20, 5, 10, 100, 30, 20],
        [1601, 100, 50, 50, 1600, 200, 201],
    ],
)
def test_determine_threshold_max_samples(
    ins,
    n_samples,
    n_remove,
    min_remove,
    min_samples,
    max_samples,
    n_live,
    expected,
    caplog,
):
    ins.min_samples = min_samples
    ins.min_remove = min_remove
    ins.max_samples = max_samples
    ins.draw_constant = True
    ins.nlive = n_live
    samples = np.empty(n_samples, dtype=[("x", "f8"), ("logL", "f8")])
    samples["logL"] = 10 * np.arange(n_samples)

    expected_log_l = samples["logL"][expected]

    ins.determine_threshold_entropy = MagicMock(return_value=n_remove)

    out = INS.determine_log_likelihood_threshold(
        ins, samples, method="entropy", q=0.8
    )

    ins.determine_threshold_entropy.assert_called_once_with(samples, q=0.8)
    assert out == expected_log_l

    if expected != n_remove:
        assert "Next level would have more than max samples" in str(
            caplog.text
        )


def test_update_log_likelihood_threshold(ins, iid):
    threshold = 10
    ins.training_samples = MagicMock(spec=OrderedSamples)
    if iid:
        ins.iid_samples = MagicMock(spec=OrderedSamples)
    INS.update_log_likelihood_threshold(ins, threshold)

    ins.training_samples.update_log_likelihood_threshold.assert_called_once_with(  # noqa
        threshold
    )
    if iid:
        ins.iid_samples.update_log_likelihood_threshold.assert_called_once_with(  # noqa
            threshold
        )



# -*- coding: utf-8 -*-
"""
Test utilities related to logging.
"""
import logging
import os
import sys
from unittest.mock import MagicMock, patch
import pytest

from nessai.utils.logging import setup_logger


def teardown_function():
    """Reset the logger after each test"""
    logger = logging.getLogger("nessai")
    logger.handlers = []
    logger.addHandler(logging.NullHandler())
    try:
        os.remove("test.log")
    except OSError:
        pass


@pytest.fixture(params=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
def log_level(request):
    """Different log levels to test."""
    return request.param


def test_setup_logger_no_label():
    """Test behaviour when label is None.

    This should NOT produce a log file.
    """
    logger = setup_logger(label=None)
    assert not any(
        [isinstance(h, logging.FileHandler) for h in logger.handlers]
    )


@pytest.mark.parametrize("output", ["logger_dir", None])
def test_setup_logger_with_label(tmp_path, output):
    """Test behaviour when label is not None.

    This should produce a log file.
    """
    if output:
        output = tmp_path / output
        output.mkdir()
    logger = setup_logger(label="test", output=output)
    if output is None:
        output = os.getcwd()
    assert os.path.exists(os.path.join(output, "test.log"))
    assert any([isinstance(h, logging.FileHandler) for h in logger.handlers])


def test_setup_logger_with_mkdir(tmp_path):
    """Assert the output directory is created if missing"""
    output = tmp_path / "logger_dir"
    setup_logger(label="test", output=output)
    assert os.path.exists(os.path.join(output, "test.log"))


@pytest.mark.parametrize(
    "log_level, value",
    [("ERROR", 40), ("WARNING", 30), ("INFO", 20), ("DEBUG", 10), (15, 15)],
)
def test_setup_logger_levels(log_level, value):
    """Verify logging levels are correctly set."""
    logger = setup_logger(log_level=log_level, label=None)
    assert all([h.level == value for h in logger.handlers])


def test_setup_logger_unknown_level():
    """Verify an error is raised for an unknown level"""
    with pytest.raises(ValueError) as excinfo:
        setup_logger(log_level="test", label=None)
    assert "log_level test not understood" in str(excinfo.value)


def test_filehandler_kwargs(tmp_path, log_level):
    """Assert filehandler kwargs are passed to the handler."""
    output = tmp_path / "logger_dir"
    handler = MagicMock(spec=logging.FileHandler)
    handler.level = 10

    class MockedFileHandler(MagicMock, logging.FileHandler):
        def __new__(cls, *args, **kwargs):
            handler(*args, **kwargs)
            return handler

    with patch("logging.FileHandler", new=MockedFileHandler):
        setup_logger(
            output=output,
            filehandler_kwargs={"mode": "w"},
            log_level=log_level,
        )
    handler.assert_called_once_with(
        os.path.join(output, "nessai.log"),
        mode="w",
    )


def test_filehandler_no_kwargs(tmp_path, log_level):
    """Assert case of no kwargs for the file handler works as intended."""
    output = tmp_path / "logger_dir"
    handler = MagicMock(spec=logging.FileHandler)
    handler.level = 10

    class MockedFileHandler(logging.FileHandler):
        def __new__(cls, *args, **kwargs):
            handler(*args, **kwargs)
            return handler

    with patch("logging.FileHandler", new=MockedFileHandler):
        setup_logger(
            output=output, filehandler_kwargs=None, log_level=log_level
        )
    handler.assert_called_once_with(
        os.path.join(output, "nessai.log"),
    )


@pytest.mark.parametrize(
    "stream, expected",
    (
        [None, None],
        ("stderr", sys.stderr),
        ("stdout", sys.stdout),
        (sys.stderr, sys.stderr),
    ),
)
def test_stream_handler_setting(tmp_path, stream, expected, log_level):
    output = tmp_path / "logger_dir"
    handler = MagicMock(spec=logging.StreamHandler)
    handler.level = 10

    class MockedStreamHandler(logging.StreamHandler):
        def __new__(cls, *args, **kwargs):
            handler(*args, **kwargs)
            return handler

    with patch("logging.StreamHandler", new=MockedStreamHandler):
        setup_logger(
            output=output, stream=stream, label=None, log_level=log_level
        )
    handler.assert_called_with(expected)


def test_stream_handler_error(tmp_path):
    """Assert an error is raised if an invalid string is passes"""
    output = tmp_path / "logger_dir"
    with pytest.raises(ValueError, match=r"Unknown stream: .*"):
        setup_logger(output=output, stream="not_a_stream")



# -*- coding: utf-8 -*-
"""
Test the rejection proposal class.
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, Mock, create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import RejectionProposal
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture
def proposal():
    return create_autospec(RejectionProposal)


def test_init(proposal):
    """Test the init method."""
    with patch(
        "nessai.proposal.rejection.AnalyticProposal.__init__"
    ) as mock_super:
        RejectionProposal.__init__(proposal, "model", poolsize=10, test=True)
    mock_super.assert_called_once_with("model", poolsize=10, test=True)
    assert proposal._checked_population is True
    assert proposal.population_acceptance is None


@pytest.mark.parametrize("N", [None, 5])
def test_draw_proposal(proposal, N):
    """Assert `model.new_point` is called with the corred number of samples."""
    points = np.array([1, 2])
    proposal.poolsize = 10
    proposal.model = Mock()
    proposal.model.new_point = Mock(return_value=points)
    samples = RejectionProposal.draw_proposal(proposal, N=N)
    np.testing.assert_array_equal(samples, points)
    if N is None:
        proposal.model.new_point.assert_called_once_with(N=10)
    else:
        proposal.model.new_point.assert_called_once_with(N=5)


def test_log_proposal(proposal):
    """Assert the correct method from the model is called"""
    x = np.array([3, 4])
    log_prob = np.array([1, 2])
    proposal.model = Mock()
    proposal.model.new_point_log_prob = Mock(return_value=log_prob)
    out = RejectionProposal.log_proposal(proposal, x)
    proposal.model.new_point_log_prob.assert_called_once_with(x)
    np.testing.assert_array_equal(out, log_prob)


@pytest.mark.parametrize("return_log_prior", [False, True])
def test_compute_weights(proposal, return_log_prior):
    """Test the compute weights method"""
    x = numpy_array_to_live_points(np.array([[1], [2], [3]]), "x")
    proposal.model = Mock()
    log_p = np.array([6, 6, 6])
    proposal.model.batch_evaluate_log_prior = Mock(return_value=log_p)
    proposal.log_proposal = Mock(return_value=np.array([3, 4, np.nan]))
    log_w = np.array([3, 2, np.nan])
    out = RejectionProposal.compute_weights(
        proposal, x, return_log_prior=return_log_prior
    )
    if return_log_prior:
        assert out[1] is log_p
        out = out[0]

    proposal.model.batch_evaluate_log_prior.assert_called_once_with(x)
    proposal.log_proposal.assert_called_once_with(x)
    np.testing.assert_array_equal(out, log_w)


@pytest.mark.parametrize("N", [None, 4])
def test_populate(proposal, N):
    """Test the populate method"""
    poolsize = 8
    if N is None:
        log_w = np.arange(poolsize)
    else:
        log_w = np.arange(N)
    x = numpy_array_to_live_points(np.random.randn(log_w.size, 1), ["x"])
    u = np.exp(log_w.copy() + 1)
    # These points will have log_u ~ -inf so corresponding samples will be
    # accepted.
    u[::2] = 1e-10
    samples = x[::2]
    log_l = np.log(np.random.rand(samples.size))
    log_prior = np.zeros(len(x))
    samples["logL"] = log_l
    proposal.poolsize = poolsize
    proposal.populated = False
    proposal.draw_proposal = Mock(return_value=x)
    proposal.compute_weights = Mock(return_value=(log_w, log_prior))
    proposal.model = Mock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )

    with patch("numpy.random.rand", return_value=u):
        RejectionProposal.populate(proposal, N=N)

    assert proposal.population_acceptance == 0.5
    assert proposal.populated is True
    assert_structured_arrays_equal(proposal.samples, samples)

    if N is None:
        N = poolsize
    proposal.draw_proposal.assert_called_once_with(N=N)

    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )
    assert sorted(proposal.indices) == list(range(samples.size))
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)


@pytest.mark.integration_test
def test_populate_integration(model):
    """Integration test for the populate method"""
    proposal = RejectionProposal(model)
    N = 500
    proposal.populate(N=N)
    assert proposal.samples.size == N
    assert proposal.populated is True



# -*- coding: utf-8 -*-
"""
Test the main sampling functions
"""
import logging
import numpy as np
import pytest
from unittest.mock import call, MagicMock

from nessai.livepoint import (
    numpy_array_to_live_points,
    parameters_to_live_point,
)
from nessai.samplers.nestedsampler import NestedSampler
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture
def live_points():
    x = numpy_array_to_live_points(np.arange(4)[:, np.newaxis], names=["x"])
    x["logL"] = np.arange(4)
    return x


@pytest.fixture
def sampler(sampler):
    sampler.state = MagicMock()
    sampler.state.logZ = -4.0
    sampler.state.info = [0.0]
    sampler.state.log_evidence_error = 0.1
    sampler.nlive = 4
    sampler.nested_samples = []
    sampler.iteration = 0
    sampler.condition = np.inf
    sampler.block_iteration = 0
    sampler.logLmax = 5
    sampler.insertion_indices = []
    sampler.accepted = 0
    sampler.rejected = 0
    sampler.block_acceptance = 0.0
    sampler.acceptance_history = []
    sampler.info_enabled = True
    sampler.debug_enabled = True
    sampler.log_on_iteration = True
    return sampler


def test_initialise(sampler):
    """Test the initialise method when being used without resuming"""
    sampler._flow_proposal = MagicMock()
    sampler._uninformed_proposal = MagicMock()
    sampler.populate_live_points = MagicMock()

    sampler._flow_proposal.initialised = False
    sampler._uninformed_proposal.initialised = False
    sampler.live_points = None
    sampler.iteration = 0
    sampler.maximum_uninformed = 10
    sampler.condition = 1.0
    sampler.tolerance = 0.1
    sampler.initialised = False
    sampler.uninformed_sampling = True

    NestedSampler.initialise(sampler)

    sampler._flow_proposal.initialise.assert_called_once()
    sampler._uninformed_proposal.initialise.assert_called_once()
    sampler.populate_live_points.assert_called_once()
    assert sampler.initialised is True
    assert sampler.proposal is sampler._uninformed_proposal


def test_initialise_resume(sampler):
    """Test the initialise method when being used after resuming.

    In this case the live points are not None
    """
    sampler._flow_proposal = MagicMock()
    sampler._uninformed_proposal = MagicMock()
    sampler.populate_live_points = MagicMock()

    sampler._flow_proposal.initialised = False
    sampler._uninformed_proposal.initialised = False
    sampler.live_points = [0.0]
    sampler.iteration = 100
    sampler.maximum_uninformed = 10
    sampler.condition = 1.0
    sampler.tolerance = 0.1
    sampler.initialised = False

    NestedSampler.initialise(sampler)

    sampler._flow_proposal.initialise.assert_called_once()
    sampler._uninformed_proposal.initialise.assert_called_once()
    sampler.populate_live_points.assert_not_called()
    assert sampler.initialised is False
    assert sampler.proposal is sampler._flow_proposal


def test_log_state(sampler, caplog):
    """Assert log state outputs to the logger"""
    caplog.set_level(logging.INFO)
    NestedSampler.log_state(sampler)
    assert "logZ:" in str(caplog.text)


def test_finalise(sampler, live_points):
    """Test the finalise method"""
    sampler.live_points = live_points
    sampler.finalised = False

    NestedSampler.finalise(sampler)

    calls = [
        call(0, nlive=4),
        call(1, nlive=3),
        call(2, nlive=2),
        call(3, nlive=1),
    ]

    sampler.state.increment.assert_has_calls(calls)
    sampler.update_state.assert_called_once_with(force=True)
    sampler.state.finalise.assert_called_once()
    assert_structured_arrays_equal(sampler.nested_samples, [*live_points])
    assert sampler.finalised is True


def test_consume_sample(sampler, live_points):
    """Test the default behaviour of consume sample"""
    sampler.live_points = live_points
    new_sample = np.squeeze(parameters_to_live_point((0.5,), ["x"]))
    new_sample["logL"] = 0.5
    sampler.yield_sample = MagicMock()
    sampler.yield_sample.return_value = iter([(1, new_sample)])

    sampler.insert_live_point = MagicMock(return_value=0)
    sampler.check_state = MagicMock()

    NestedSampler.consume_sample(sampler)

    sampler.insert_live_point.assert_called_once_with(new_sample)
    sampler.check_state.assert_not_called()

    assert_structured_arrays_equal(sampler.nested_samples, [live_points[0]])
    assert sampler.logLmin == 0.0
    assert sampler.accepted == 1
    assert sampler.block_acceptance == 1.0
    assert sampler.acceptance_history == [1.0]
    assert sampler.mean_block_acceptance == 1.0
    assert sampler.insertion_indices == [0]


def test_consume_sample_reject(sampler, live_points):
    """Test the default behaviour of consume sample"""
    sampler.live_points = live_points
    reject_sample = parameters_to_live_point((-0.5,), ["x"])
    reject_sample["logL"] = -0.5
    new_sample = np.squeeze(parameters_to_live_point((0.5,), ["x"]))
    new_sample["logL"] = 0.5
    sampler.yield_sample = MagicMock()
    sampler.yield_sample.return_value = iter(
        [(1, reject_sample), (1, new_sample)]
    )

    sampler.insert_live_point = MagicMock(return_value=0)
    sampler.check_state = MagicMock()

    NestedSampler.consume_sample(sampler)

    sampler.insert_live_point.assert_called_once_with(new_sample)
    sampler.check_state.assert_called_once()

    assert_structured_arrays_equal(sampler.nested_samples, [live_points[0]])
    assert sampler.logLmin == 0.0
    assert sampler.rejected == 1
    assert sampler.accepted == 1
    assert sampler.block_acceptance == 0.5
    assert sampler.acceptance_history == [0.5]
    assert sampler.mean_block_acceptance == 0.5
    assert sampler.insertion_indices == [0]


@pytest.mark.parametrize(
    "config",
    [
        {
            "tolerance": 0.1,
            "condition": 0.01,
            "call_finalise": True,
            "call_while": False,
            "close_pool": True,
        },
        {
            "iteration": 10,
            "max_iteration": 10,
            "call_finalise": False,
            "call_while": True,
            "close_pool": False,
        },
    ],
)
def test_nested_sampling_loop(sampler, config):
    """Test the main nested sampling loop.

    This is hard to test because of while loop.
    """
    sampler.prior_sampling = False
    sampler.initialised = False
    sampler.condition = config.get("condition", 0.5)
    sampler.tolerance = config.get("tolerance", 0.1)
    sampler._close_pool = config.get("close_pool", True)
    sampler.max_iteration = config.get("max_iteration")
    sampler.iteration = config.get("iteration", 0)
    sampler.sampling_time = 0.0
    sampler.training_time = 0.0
    sampler.proposal_population_time = 0.0
    sampler.likelihood_calls = 1
    sampler.nested_samples = [1, 2]

    sampler.close_pool = MagicMock()

    sampler.proposal = MagicMock()
    sampler.proposal.pool = True
    sampler.proposal.logl_eval_time.total_seconds = MagicMock()

    sampler.finalised = False
    sampler.finalise = MagicMock()

    sampler.check_resume = MagicMock()
    sampler.check_insertion_indices = MagicMock()
    sampler.checkpoint = MagicMock()

    logZ, samples = NestedSampler.nested_sampling_loop(sampler)

    assert logZ == sampler.state.logZ
    assert samples.tolist() == [1, 2]

    sampler.initialise.assert_called_once_with(live_points=True)
    sampler.check_resume.assert_called_once()

    # If the code entered the while loop, check the functions where called
    if config.get("call_while", True):
        if config.get("iteration", 0):
            sampler.update_state.call_count == 2
        else:
            sampler.update_state.assert_called_once()
        sampler.consume_sample.assert_called_once()
        sampler.check_state.assert_called_once()
        sampler.periodically_log_state.assert_called_once()
    else:
        sampler.check_state.assert_not_called()
        sampler.consume_sample.assert_not_called()
        sampler.update_state.assert_not_called()
        sampler.periodically_log_state.assert_not_called()

    if config.get("call_finalise"):
        sampler.finalise.assert_called_once()
    else:
        sampler.finalise.assert_not_called()

    sampler.check_insertion_indices.assert_called_once_with(rolling=False)
    sampler.checkpoint.assert_called_once_with(periodic=True, force=True)

    if sampler._close_pool:
        sampler.close_pool.assert_called_once()
    else:
        sampler.close_pool.assert_not_called()


@pytest.mark.parametrize("close_pool", [False, True])
def test_nested_sampling_loop_prior_sampling(sampler, close_pool):
    """Test the nested sampling loop for prior sampling"""
    sampler.initialised = False
    sampler.nested_samples = sampler.model.new_point(10)
    sampler.prior_sampling = True
    sampler._close_pool = close_pool
    sampler.close_pool = MagicMock()
    sampler.finalise = MagicMock()
    sampler.log_evidence = -5.99

    evidence, samples = NestedSampler.nested_sampling_loop(sampler)
    sampler.initialise.assert_called_once_with(live_points=True)
    if close_pool:
        sampler.close_pool.assert_called_once()
    else:
        sampler.close_pool.assert_not_called()
    sampler.finalise.assert_called_once()
    assert_structured_arrays_equal(samples, sampler.nested_samples)
    assert evidence == -5.99



# -*- coding: utf-8 -*-
"""Test the base nested sampler"""
import datetime
import os
import pickle
import pytest
import time
from unittest.mock import MagicMock, create_autospec, patch

from nessai.samplers.base import BaseNestedSampler


@pytest.fixture
def sampler():
    obj = create_autospec(BaseNestedSampler)
    obj.model = MagicMock()
    return obj


@pytest.mark.parametrize("checkpoint_on_iteration", [False, True])
def test_init(sampler, checkpoint_on_iteration):
    """Assert the init sets the correct attributes and calls the correct
    methods.
    """
    model = MagicMock()
    model.configure_pool = MagicMock()
    model.verify_model = MagicMock()
    nlive = 100
    output = "./"
    seed = 190425
    checkpointing = True
    checkpoint_interval = 10
    log_on_iteration = True
    logging_interval = 10
    resume_file = "test.pkl"
    plot = False
    n_pool = 2
    pool = MagicMock()

    sampler.configure_random_seed = MagicMock()
    sampler.configure_output = MagicMock()
    sampler.configure_periodic_logging = MagicMock()

    BaseNestedSampler.__init__(
        sampler,
        model,
        nlive,
        output=output,
        seed=seed,
        checkpointing=checkpointing,
        checkpoint_interval=checkpoint_interval,
        checkpoint_on_iteration=checkpoint_on_iteration,
        log_on_iteration=log_on_iteration,
        logging_interval=logging_interval,
        resume_file=resume_file,
        plot=plot,
        n_pool=n_pool,
        pool=pool,
    )

    model.verify_model.assert_called_once()
    model.configure_pool.assert_called_once_with(pool=pool, n_pool=n_pool)

    sampler.configure_random_seed.assert_called_once_with(seed)
    sampler.configure_output.assert_called_once_with(
        output, resume_file=resume_file
    )
    sampler.configure_periodic_logging.assert_called_once_with(
        logging_interval,
        log_on_iteration,
    )

    assert sampler.nlive == nlive
    assert sampler.plot == plot
    assert sampler.checkpointing == checkpointing
    assert sampler.checkpoint_interval == checkpoint_interval
    assert sampler.checkpoint_on_iteration == checkpoint_on_iteration
    assert sampler.live_points is None
    assert sampler.iteration == 0
    assert sampler.finalised is False
    assert sampler.history is None


def test_likelihood_evaluation_time(sampler):
    time = datetime.timedelta(seconds=4)
    sampler.model.likelihood_evaluation_time = time
    out = BaseNestedSampler.likelihood_evaluation_time.__get__(sampler)
    assert out is time


def test_current_sampling_time(sampler, wait):
    """Test the current sampling time"""
    sampler.finalised = False
    sampler.sampling_time = datetime.timedelta(seconds=10)
    sampler.sampling_start_time = datetime.datetime.now()
    wait()
    t = BaseNestedSampler.current_sampling_time.__get__(sampler)
    assert t.total_seconds() > 10.0


def test_total_likelihood_evaluations(sampler):
    """Check likelihood calls from model are returned"""
    sampler.model = MagicMock()
    sampler.model.likelihood_evaluations = 10
    assert (
        BaseNestedSampler.total_likelihood_evaluations.__get__(sampler) == 10
    )


def test_likelihood_calls(sampler):
    """Check likelihood calls from model are returned"""
    sampler.model = MagicMock()
    sampler.model.likelihood_evaluations = 10
    assert BaseNestedSampler.likelihood_calls.__get__(sampler) == 10


def test_current_sampling_time_finalised(sampler):
    """Test the current sampling time if the sampling has been finalised"""
    sampler.finalised = True
    sampler.sampling_time = 10
    assert BaseNestedSampler.current_sampling_time.__get__(sampler) == 10


def test_posterior_effective_sample_suze(sampler):
    """Assert an error is raised"""
    with pytest.raises(NotImplementedError):
        BaseNestedSampler.posterior_effective_sample_size.__get__(sampler)


@patch("numpy.random.seed")
@patch("torch.manual_seed")
def test_set_random_seed(mock1, mock2, sampler):
    """Test the correct functions are called when setting the random seed"""
    BaseNestedSampler.configure_random_seed(sampler, 150914)
    mock1.assert_called_once_with(150914)
    mock2.assert_called_once_with(seed=150914)
    assert sampler.seed == 150914


@patch("numpy.random.seed")
@patch("torch.manual_seed")
@patch("numpy.random.randint", return_value=10)
def test_no_random_seed(mock_int, mock1, mock2, sampler):
    """Assert a random seed is set if seed=None"""
    BaseNestedSampler.configure_random_seed(sampler, None)
    mock_int.assert_called_once()
    mock1.assert_called_once_with(10)
    mock2.assert_called_once_with(seed=10)
    assert sampler.seed == 10


@patch("numpy.random.seed")
@patch("torch.manual_seed")
def test_no_random_seed_numpy(mock1, mock2, sampler):
    """Assert a random seed is set if seed=None"""
    BaseNestedSampler.configure_random_seed(sampler, None)
    mock1.assert_called_once()
    mock2.assert_called_once()


def test_configure_output(sampler, tmpdir):
    """Test setting up the output directories"""
    p = tmpdir.mkdir("outputs")
    sampler.plot = False
    path = os.path.join(p, "tests")
    BaseNestedSampler.configure_output(sampler, path)
    assert os.path.exists(path)
    assert sampler.resume_file == os.path.join(
        path, "nested_sampler_resume.pkl"
    )


def test_configure_output_none(sampler, tmpdir):
    """Test setting up the output directories if the output is None"""
    p = tmpdir.mkdir("outputs")
    sampler.plot = False
    path = os.path.join(p, "test_cwd")
    with patch("os.getcwd", return_value=path) as mock:
        BaseNestedSampler.configure_output(sampler, None)

    mock.assert_called_once()
    assert sampler.output == path


def test_configure_output_w_resume(sampler, tmpdir):
    """Test output configuration with a specified resume file"""
    p = tmpdir.mkdir("outputs")
    sampler.plot = False
    path = os.path.join(p, "tests")
    BaseNestedSampler.configure_output(sampler, path, "resume.pkl")
    assert sampler.resume_file == os.path.join(path, "resume.pkl")


def test_configure_periodic_logging_time(sampler):
    """Assert the perdiodic logging is correctly configured"""
    with patch("time.time", return_value=10) as mock:
        BaseNestedSampler.configure_periodic_logging(
            sampler,
            20,
            False,
        )
    mock.assert_called_once()
    assert sampler._last_log == 10
    assert sampler.logging_interval == 20
    assert sampler.log_on_iteration is False


@pytest.mark.parametrize("interval, expected", [(50, 50), (None, 100)])
def test_configure_periodic_logging_interval(sampler, interval, expected):
    """Assert the perdiodic logging is correctly configured"""
    sampler.nlive = 100
    BaseNestedSampler.configure_periodic_logging(
        sampler,
        interval,
        True,
    )
    assert sampler._last_log == 0
    assert sampler.logging_interval == expected
    assert sampler.log_on_iteration is True


def test_configure_periodic_logging_all_false(sampler):
    """Assert log on iteration is enabled if both inputs are false"""
    sampler.nlive = 100
    BaseNestedSampler.configure_periodic_logging(sampler, None, False)
    assert sampler.log_on_iteration is True
    assert sampler._last_log == 0
    assert sampler.logging_interval == 100


def test_log_state(sampler):
    """Assert a NotImplementedError is raised"""
    with pytest.raises(NotImplementedError):
        BaseNestedSampler.log_state(sampler)


@pytest.mark.parametrize("interval", [20, 40])
def test_periodically_log_state_time(sampler, interval):
    """Assert log_state is called sufficient time has elapsed.

    Set such that 30 seconds have elapsed since the last log.
    """
    now = time.time()
    elapsed = 30
    sampler._last_log = now - elapsed
    sampler.logging_interval = interval
    sampler.log_on_iteration = False
    with patch("time.time", return_value=now):
        BaseNestedSampler.periodically_log_state(sampler)
    # if more time has passed, then should log
    if elapsed > interval:
        sampler.log_state.assert_called_once()
        assert sampler._last_log == now
    else:
        sampler.log_state.assert_not_called()
        assert sampler._last_log == (now - elapsed)


@pytest.mark.parametrize("interval", [20, 40])
def test_periodically_log_state_iteration(sampler, interval):
    """Assert log_state is called sufficient iterations have elapsed.

    Set such that 30 iterations have elapsed since the last log.
    """
    elapsed = 30
    sampler.iteration = 120
    sampler._last_log = sampler.iteration - elapsed
    sampler.logging_interval = interval
    sampler.log_on_iteration = True
    BaseNestedSampler.periodically_log_state(sampler)
    # if more iterations have passed, then should log
    if elapsed > interval:
        sampler.log_state.assert_called_once()
        assert sampler._last_log == sampler.iteration
    else:
        sampler.log_state.assert_not_called()
        assert sampler._last_log == (sampler.iteration - elapsed)


@pytest.mark.parametrize("periodic", [False, True])
@pytest.mark.parametrize("no_history", [False, True])
def test_checkpoint_iteration(sampler, wait, periodic, no_history):
    """Test checkpointing method on iterations.

    Make sure a file is produced and that the sampling time is updated.
    Also checks to make sure that the iteration is recorded when periodic=False
    """
    if no_history:
        sampler.history = None
    else:
        sampler.history = dict(checkpoint_iterations=[10])
    sampler.checkpoint_on_iteration = True
    sampler.checkpoint_interval = 10
    sampler.checkpoint_callback = None
    sampler._last_checkpoint = 0
    sampler.iteration = 20
    now = datetime.datetime.now()
    sampler.sampling_start_time = now
    sampler.sampling_time = datetime.timedelta()
    sampler.resume_file = "test.pkl"

    with patch("nessai.samplers.base.safe_file_dump") as sfd_mock:
        wait()
        BaseNestedSampler.checkpoint(sampler, periodic=periodic)

    sfd_mock.assert_called_once_with(
        sampler, sampler.resume_file, pickle, save_existing=True
    )

    assert sampler.sampling_start_time > now
    assert sampler.sampling_time.total_seconds() > 0.0

    if periodic:
        if not no_history:
            assert sampler.history["checkpoint_iterations"] == [10]
        else:
            assert sampler.history is None
        assert sampler._last_checkpoint == 20
    else:
        if not no_history:
            assert sampler.history["checkpoint_iterations"] == [10, 20]
        else:
            assert sampler.history is None


def test_checkpoint_time(sampler, wait):
    """Test checkpointing method based on time interval

    Make sure a file is produced and that the sampling time is updated.
    """
    now = datetime.datetime.now()
    sampler.checkpoint_iterations = [10]
    sampler.checkpoint_on_iteration = False
    sampler.checkpoint_interval = 15 * 60
    sampler.checkpoint_callback = None
    sampler.sampling_start_time = now - datetime.timedelta(minutes=32)
    sampler._last_checkpoint = now - datetime.timedelta(minutes=16)
    sampler.iteration = 20
    sampler.sampling_time = datetime.timedelta()
    sampler.resume_file = "test.pkl"

    with patch("nessai.samplers.base.safe_file_dump") as sfd_mock, patch(
        "datetime.datetime", return_value=now
    ) as mock_datetime:
        wait()
        mock_datetime.now.return_value = now
        mock_datetime.side_effect = lambda *args, **kw: datetime.datetime(
            *args, **kw
        )
        BaseNestedSampler.checkpoint(sampler, periodic=True)

    sfd_mock.assert_called_once_with(
        sampler, sampler.resume_file, pickle, save_existing=True
    )
    assert sampler._last_checkpoint is now


def test_checkpoint_periodic_skipped_iteration(sampler):
    """Assert the sampler does not checkpoint if the criterion is not met"""
    sampler.checkpoint_on_iteration = True
    sampler.iteration = 10
    sampler._last_checkpoint = 9
    sampler.checkpoint_interval = 10
    sampler.checkpoint_callback = None
    with patch("nessai.samplers.base.safe_file_dump") as sfd_mock:
        BaseNestedSampler.checkpoint(sampler, periodic=True)
    sfd_mock.assert_not_called()


def test_checkpoint_periodic_skipped_time(sampler):
    """Assert the sampler does not checkpoint if the criterion is not met"""
    sampler.checkpoint_on_iteration = False
    sampler.iteration = 10
    sampler._last_checkpoint = datetime.datetime.now()
    sampler.checkpoint_interval = 600
    sampler.checkpoint_callback = None
    with patch("nessai.samplers.base.safe_file_dump") as sfd_mock:
        BaseNestedSampler.checkpoint(sampler, periodic=True)
    sfd_mock.assert_not_called()


def test_checkpoint_force(sampler):
    """Assert the sampler checkpoints if force=True"""
    now = datetime.datetime.now()
    sampler.sampling_start_time = now - datetime.timedelta(minutes=32)
    sampler.sampling_time = datetime.timedelta()
    sampler.resume_file = "test.pkl"
    sampler.checkpoint_callback = None
    with patch("nessai.samplers.base.safe_file_dump") as sfd_mock:
        BaseNestedSampler.checkpoint(sampler, periodic=True, force=True)
    sfd_mock.assert_called_once_with(
        sampler, sampler.resume_file, pickle, save_existing=True
    )


def test_checkpoint_callback(sampler):
    """Assert the checkpoint callback is used"""

    callback = MagicMock()

    sampler.history = dict(checkpoint_iterations=[10])
    sampler.checkpoint_on_iteration = True
    sampler.checkpoint_interval = 10
    sampler.checkpoint_callback = callback
    sampler._last_checkpoint = 0
    sampler.iteration = 20
    now = datetime.datetime.now()
    sampler.sampling_start_time = now
    sampler.sampling_time = datetime.timedelta()

    BaseNestedSampler.checkpoint(sampler)
    callback.assert_called_once_with(sampler)


def test_nested_sampling_loop(sampler):
    """Assert an error is raised"""
    with pytest.raises(NotImplementedError):
        BaseNestedSampler.nested_sampling_loop(sampler)


def test_close_pool(sampler):
    """Assert the method in the model is called with the correct code"""
    sampler.model.close_pool = MagicMock()
    BaseNestedSampler.close_pool(sampler, 2)
    sampler.model.close_pool.assert_called_once_with(code=2)


def test_resume_from_pickled_sampler(model):
    """Test the resume from pickled sampler method"""
    obj = MagicMock()
    obj.model = None
    obj._previous_likelihood_evaluations = 3
    obj._previous_likelihood_evaluation_time = 4.0

    model.likelihood_evaluations = 1
    model.likelihood_evaluation_time = datetime.timedelta(seconds=2)

    out = BaseNestedSampler.resume_from_pickled_sampler(obj, model)

    assert out.model == model
    assert out.model.likelihood_evaluations == 4
    assert out.model.likelihood_evaluation_time.total_seconds() == 6


def test_resume(model):
    """Test the resume method"""
    obj = MagicMock()
    pickle_out = MagicMock()

    with patch("pickle.load", return_value=obj) as mock_pickle, patch(
        "builtins.open"
    ), patch(
        "nessai.samplers.base.BaseNestedSampler.resume_from_pickled_sampler",
        return_value=pickle_out,
    ) as mock_resume:
        out = BaseNestedSampler.resume("test.pkl", model)

    assert out is pickle_out
    mock_pickle.assert_called_once()
    mock_resume.assert_called_once_with(obj, model)


def test_get_result_dictionary(sampler):
    """Assert the correct dictionary is returned"""
    sampler.seed = 170817
    sampler.sampling_time = datetime.timedelta(seconds=4)
    sampler.likelihood_evaluation_time = datetime.timedelta(seconds=2)
    sampler.model.truth = 1.0
    sampler.model.likelihood_evaluations = 10
    sampler.history = None

    d = BaseNestedSampler.get_result_dictionary(sampler)

    assert d["seed"] == sampler.seed
    assert d["sampling_time"] == 4
    assert d["total_likelihood_evaluations"] == 10
    assert d["likelihood_evaluation_time"] == 2
    assert d["truth"] == 1.0
    assert "history" in d


def test_getstate(sampler):
    """Assert the model is deleted and the evaluations are storred"""
    sampler.model.likelihood_evaluations = 10
    sampler.model.likelihood_evaluation_time = datetime.timedelta(seconds=4)
    d = BaseNestedSampler.__getstate__(sampler)
    assert "model" not in d
    assert d["_previous_likelihood_evaluations"] == 10
    assert d["_previous_likelihood_evaluation_time"] == 4


def test_initialise_history(sampler):
    sampler.history = None
    BaseNestedSampler.initialise_history(sampler)
    assert isinstance(sampler.history, dict)
    assert "sampling_time" in sampler.history
    assert "likelihood_evaluations" in sampler.history


def test_initialise_history_skip(sampler, caplog):
    caplog.set_level("DEBUG")
    sampler.history = {}
    BaseNestedSampler.initialise_history(sampler)
    assert "already initialised" in str(caplog.text)


def test_update_history(sampler):
    sampler.total_likelihood_evaluations = 20
    sampler.current_sampling_time = datetime.timedelta(seconds=2)
    sampler.history = dict(likelihood_evaluations=[10], sampling_time=[1])
    BaseNestedSampler.update_history(sampler)
    assert sampler.history["likelihood_evaluations"] == [10, 20]
    assert sampler.history["sampling_time"] == [1, 2]



# -*- coding: utf-8 -*-
"""
Functions related to computing the posterior samples.
"""
import logging

import numpy as np
from scipy.special import logsumexp

from .evidence import logsubexp, log_integrate_log_trap
from .utils.stats import effective_sample_size

logger = logging.getLogger(__name__)


def compute_weights(samples, nlive, expectation="logt"):
    """
    Returns the log-evidence and log-weights for the log-likelihood samples
    assumed to the result of nested sampling with nlive live points

    Parameters
    ----------
    samples : array_like
        Log-likelihood samples.
    nlive : Union[int, array_like]
        Number of live points used in nested sampling.
    expectation : str, {logt, t}
        Method used to compute the expectation value for the shrinkage t.
        Choose between log <t> or <log t>. Defaults to <log t>.

    Returns
    -------
    float
        The computed log-evidence.
    array_like
        Array of computed weights (already normalised by the log-evidence).
    """
    samples = np.asarray(samples)

    if isinstance(nlive, (int, float)):
        nlive_per_iteration = nlive * np.ones_like(samples)
        nlive_per_iteration[-nlive:] = np.arange(nlive, 0, -1, dtype=float)
    else:
        if len(nlive) != len(samples):
            raise ValueError("nlive and samples are different lengths")
        nlive_per_iteration = nlive.copy()

    if expectation.lower() == "logt":
        logt = -1.0 / nlive_per_iteration
    elif expectation.lower() == "t":
        logt = -np.log1p(1.0 / nlive_per_iteration)
    else:
        raise ValueError(f"Expectation must be t or logt, got: {expectation}")

    # One point at X=1 and X=0
    n_vols = len(samples) + 2
    log_vols = np.zeros(n_vols)
    log_vols[1:-1] = np.cumsum(logt)
    log_vols[-1] = np.NINF

    # First point represents the entire prior
    # Last point represents X=0 and assume max(L) = L[-1]
    log_likelihoods = np.concatenate(
        [np.array([np.NINF]), samples, np.array([samples[-1]])]
    )

    log_evidence = log_integrate_log_trap(log_likelihoods, log_vols)

    log_w = logsubexp(log_vols[:-1], log_vols[1:])

    log_post_w = log_likelihoods[1:-1] + log_w[:-1]
    log_post_w -= log_evidence

    return log_evidence, log_post_w


def draw_posterior_samples(
    nested_samples,
    nlive=None,
    n=None,
    log_w=None,
    method="rejection_sampling",
    return_indices=False,
    expectation="logt",
):
    """Draw posterior samples given the nested samples.

    Requires either the posterior weights or the number of live points.

    Parameters
    ----------
    nested_samples : structured array
        Array of nested samples.
    nlive : int, optional
        Number of live points used during nested sampling. Either this
        arguments or log_w must be specified.
    n : int, optional
        Number of samples to draw. Only used for importance sampling. If not
        specified, the effective sample size is used instead.
    log_w : array_like, optional
        Array of posterior weights. If specified the weights are not computed
        and these weights are used instead.
    method : str
        Method for drawing the posterior samples. Choose from

            - :code:`'rejection_sampling'`
            - :code:`'multinomial_resampling'`
            - :code:`'importance_sampling'` (same as multinomial)
    return_indices : bool
        If true return the indices of the accepted samples.
    expectation : str, {logt, t}
        Method used to compute the expectation value for the shrinkage t.
        Choose between log <t> or <log t>. Defaults to <log t>. Only used when
        :code:`log_w` is not specified.

    Returns
    -------
    posterior_samples : numpy.ndarray
        Samples from the posterior distribution.
    indices : numpy.ndarray
        Indices of the accepted samples in the original nested samples.
        Only returned if :code:`return_indices` is True.

    Raises
    ------
    ValueError
        If the chosen method is not a valid method.
    """
    nested_samples = np.asarray(nested_samples)
    if log_w is None:
        _, log_w = compute_weights(
            nested_samples["logL"], nlive, expectation=expectation
        )
    else:
        log_w = np.asarray(log_w)
    ess = effective_sample_size(log_w)
    logger.info(f"Effective sample size: {ess:.1f}")
    if method == "rejection_sampling":
        logger.info("Producing posterior samples using rejection sampling")
        if n is not None:
            logger.warning(
                "Number of samples cannot be specified for rejection sampling"
            )
        log_w = log_w - np.max(log_w)
        n_expected = np.exp(logsumexp(log_w))
        logger.info(f"Expect {n_expected} samples from rejection sampling")
        log_u = np.log(np.random.rand(nested_samples.size))
        indices = np.where(log_w > log_u)[0]
        samples = nested_samples[indices]
    elif method in ["importance_sampling", "multinomial_resampling"]:
        logger.info("Producing posterior samples using multinomial resampling")
        if n is None:
            n = int(ess)
        log_w = log_w - logsumexp(log_w)
        indices = np.random.choice(
            nested_samples.size, size=n, p=np.exp(log_w), replace=True
        )
        samples = nested_samples[indices]
    else:
        raise ValueError(
            f"Unknown method of drawing posterior samples: {method}"
        )

    if return_indices:
        return samples, indices
    else:
        return samples



# -*- coding: utf-8 -*-
"""
Code related to the implementation of normalising flows and some complete
implementations.
"""
from .maf import MaskedAutoregressiveFlow
from .nsf import NeuralSplineFlow
from .realnvp import RealNVP
from .utils import configure_model, reset_weights, reset_permutations

__all__ = [
    "MaskedAutoregressiveFlow",
    "NeuralSplineFlow",
    "RealNVP",
    "configure_model",
    "reset_weights",
    "reset_permutations",
]



# -*- coding: utf-8 -*-
"""
Tests for the FlowSampler class.
"""
import json
import logging
import os
import pickle
import signal
import sys
import time
from threading import Thread
from unittest.mock import MagicMock, call, create_autospec, patch


import h5py
import pytest
from nessai import config as nessai_config
from nessai.evidence import _NSIntegralState
from nessai.flowsampler import FlowSampler
from nessai.samplers import NestedSampler
from nessai.livepoint import numpy_array_to_live_points
import numpy as np


@pytest.fixture()
def flow_sampler():
    sampler = create_autospec(FlowSampler)
    sampler.importance_nested_sampler = False
    return sampler


@pytest.fixture(autouse=True)
def reset_handlers():
    """Reset signal handling after each test"""
    yield
    # Just in case the tests were ran on Windows
    try:
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
    except AttributeError:
        print("Cannot set signal attributes on this system.")


@pytest.fixture()
def names():
    return ["x", "y"]


@pytest.fixture()
def posterior_samples(names):
    return numpy_array_to_live_points(np.random.randn(10, len(names)), names)


@pytest.fixture()
def nested_samples(names):
    return numpy_array_to_live_points(np.random.randn(20, len(names)), names)


def test_nested_samples_final(flow_sampler):
    flow_sampler._final_samples = [
        1,
    ]
    flow_sampler._nested_samples = [
        2,
    ]
    assert (
        FlowSampler.nested_samples.__get__(flow_sampler)
        is flow_sampler._final_samples
    )


def test_nested_samples_ns(flow_sampler):
    flow_sampler._final_samples = None
    flow_sampler._nested_samples = [
        2,
    ]
    assert (
        FlowSampler.nested_samples.__get__(flow_sampler)
        is flow_sampler._nested_samples
    )


def test_check_resume_data_only(flow_sampler):
    assert (
        FlowSampler.check_resume(
            flow_sampler, resume_file=None, resume_data=object()
        )
        is True
    )


def test_check_resume_neither(flow_sampler):
    assert (
        FlowSampler.check_resume(
            flow_sampler, resume_file=None, resume_data=None
        )
        is False
    )


def test_check_resume_files_do_not_exist(flow_sampler, tmp_path):
    flow_sampler.output = tmp_path / "test"
    assert (
        FlowSampler.check_resume(
            flow_sampler, resume_file="test.pkl", resume_data=None
        )
        is False
    )


@pytest.mark.parametrize("resume", [False, True])
@pytest.mark.parametrize("use_ins", [False, True])
def test_init_no_resume_file(flow_sampler, tmp_path, resume, use_ins):
    """Test the init method when there is no run to resume from"""

    integration_model = MagicMock()
    output = tmp_path / "init"
    output.mkdir()
    output = str(output)
    resume = resume
    exit_code = 131
    pytorch_threads = 2
    resume_file = "test.pkl"
    kwargs = dict(
        nlive=1000,
    )
    close_pool = True

    sampler_class = "ImportanceNestedSampler" if use_ins else "NestedSampler"

    flow_sampler.save_kwargs = MagicMock()
    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch(
        f"nessai.flowsampler.{sampler_class}", return_value="ns"
    ) as mock, patch("nessai.flowsampler.configure_threads") as mock_threads:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            resume=resume,
            exit_code=exit_code,
            pytorch_threads=pytorch_threads,
            resume_file=resume_file,
            close_pool=close_pool,
            importance_nested_sampler=use_ins,
            **kwargs,
        )

    mock_threads.assert_called_once_with(
        pytorch_threads=pytorch_threads,
    )

    mock.assert_called_once_with(
        integration_model,
        output=os.path.join(output, ""),
        resume_file=resume_file,
        close_pool=False,
        **kwargs,
    )

    assert flow_sampler.ns == "ns"

    flow_sampler.save_kwargs.assert_called_once_with(kwargs)


def test_resume_from_resume_data(flow_sampler, model, tmp_path):
    """Test for resume from data"""
    output = tmp_path / "test"
    data = object()
    flow_sampler.check_resume = MagicMock(return_value=True)
    flow_sampler._resume_from_data = MagicMock()
    FlowSampler.__init__(
        flow_sampler, model, output=output, resume_data=data, resume=True
    )
    flow_sampler._resume_from_data.assert_called_once_with(
        NestedSampler,
        resume_data=data,
        model=model,
        weights_path=None,
        flow_config=None,
        checkpoint_callback=None,
    )


def test_resume_from_resume_file(flow_sampler, model, tmp_path):
    """Test for resume from data"""
    output = tmp_path / "test"
    resume_file = "resume.pkl"
    flow_sampler.check_resume = MagicMock(return_value=True)
    flow_sampler._resume_from_file = MagicMock()
    FlowSampler.__init__(
        flow_sampler,
        model,
        output=output,
        resume_data=None,
        resume=True,
        resume_file=resume_file,
    )
    flow_sampler._resume_from_file.assert_called_once_with(
        NestedSampler,
        resume_file=resume_file,
        model=model,
        weights_path=None,
        flow_config=None,
        checkpoint_callback=None,
    )


def test_init_eps(flow_sampler, tmp_path):
    initial_eps = nessai_config.general.eps

    model = MagicMock()
    output = tmp_path / "init"
    output.mkdir()
    output = str(output)
    eps = 1e-4

    flow_sampler.save_kwargs = MagicMock()

    with patch("nessai.flowsampler.NestedSampler", return_value="ns"), patch(
        "nessai.flowsampler.configure_threads"
    ):
        FlowSampler.__init__(flow_sampler, model, output=output, eps=eps)

    assert nessai_config.general.eps == eps
    assert nessai_config.general.eps != initial_eps
    nessai_config.general.eps = initial_eps


def test_disable_vectorisation(flow_sampler, tmp_path):
    """Assert vectorisation is disabled"""
    output = tmp_path / "test"
    output.mkdir()

    integration_model = MagicMock()
    integration_model.allow_vectorised = True
    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch("nessai.flowsampler.NestedSampler") as mock:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            disable_vectorisation=True,
        )
    mock.assert_called_once()
    input_integration_model = mock.call_args[0][0]
    assert input_integration_model.allow_vectorised is False


def test_likelihood_chunksize(flow_sampler, tmp_path):
    """Assert the likelihood chunksize is set."""
    output = tmp_path / "test"
    output.mkdir()

    integration_model = MagicMock()
    integration_model.likelihood_chunksize = None
    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch("nessai.flowsampler.NestedSampler") as mock:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            likelihood_chunksize=100,
        )
    mock.assert_called_once()
    input_integration_model = mock.call_args[0][0]
    assert input_integration_model.likelihood_chunksize == 100


def test_allow_multi_valued_likelihood(flow_sampler, tmp_path):
    """Assert allow_multi_valued_likelihood is true"""
    output = tmp_path / "test"
    output.mkdir()

    integration_model = MagicMock()
    integration_model.allow_multi_value_likelihood = False
    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch("nessai.flowsampler.NestedSampler") as mock:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            allow_multi_valued_likelihood=True,
        )
    mock.assert_called_once()
    input_integration_model = mock.call_args[0][0]
    assert input_integration_model.allow_multi_valued_likelihood is True


@pytest.mark.parametrize("value", [False, True])
def test_parallelise_prior(flow_sampler, tmp_path, value):
    """Assert parallise_prior is the correct value"""
    output = tmp_path / "test"
    output.mkdir()

    integration_model = MagicMock()
    integration_model.parallelise_prior = None
    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch("nessai.flowsampler.NestedSampler") as mock:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            parallelise_prior=value,
        )
    mock.assert_called_once()
    input_integration_model = mock.call_args[0][0]
    assert input_integration_model.parallelise_prior is value


@pytest.mark.parametrize(
    "test_old, error",
    [(False, None), (True, RuntimeError), (True, FileNotFoundError)],
)
@pytest.mark.integration_test
def test_init_resume(tmp_path, test_old, error):
    """Test the init method when the sampler should resume.

    Tests the case where the first file works and the case where the first
    file fails but the old method works.
    """
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    resume = True
    exit_code = 131
    pytorch_threads = 2
    resume_file = "test.pkl"
    weights_file = "integration_model.pt"
    flow_config = dict(lr=0.1)

    if test_old:
        expected_rf = output / (resume_file + ".old")
        side_effect = [error, "ns"]
    else:
        expected_rf = output / resume_file
        side_effect = ["ns"]
    expected_rf.write_text("contents")

    output = str(output)
    expected_rf = str(expected_rf)

    kwargs = dict(
        nlive=1000,
        flow_config=flow_config,
    )

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ) as mock_resume, patch(
        "nessai.flowsampler.configure_threads"
    ) as mock_threads:
        fs = FlowSampler(
            integration_model,
            output=output,
            resume=resume,
            exit_code=exit_code,
            pytorch_threads=pytorch_threads,
            resume_file=resume_file,
            weights_file=weights_file,
            **kwargs,
        )

    mock_threads.assert_called_once_with(
        pytorch_threads=pytorch_threads,
    )

    mock_resume.assert_called_with(
        expected_rf,
        integration_model,
        flow_config=flow_config,
        weights_path=weights_file,
        checkpoint_callback=None,
    )

    assert fs.ns == "ns"


def test_resume_error_cannot_resume(flow_sampler, tmp_path):
    """Assert an error is raised if neither file loads"""
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    resume_file = "test.pkl"
    expected_rf = output / (resume_file + ".old")
    expected_rf.write_text("contents")
    side_effect = [RuntimeError, RuntimeError]
    output = str(output)
    expected_rf = str(expected_rf)

    assert os.path.exists(expected_rf)

    flow_sampler.output = output

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ), patch("nessai.flowsampler.configure_threads"), pytest.raises(
        RuntimeError, match=r"Could not resume sampler with error:"
    ):
        FlowSampler._resume_from_file(
            flow_sampler,
            NestedSampler,
            integration_model,
            resume_file,
            weights_path=None,
            flow_config=None,
        )


@pytest.mark.parametrize(
    "test_old, error",
    [(False, None), (True, RuntimeError), (True, FileNotFoundError)],
)
def test_resume_from_file(flow_sampler, tmp_path, test_old, error):
    """Assert the sampler can be resumed"""
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    resume_file = "test.pkl"
    ns = object()

    if test_old:
        expected_rf = output / (resume_file + ".old")
        side_effect = [error, ns]
    else:
        expected_rf = output / resume_file
        side_effect = [ns]
    expected_rf.write_text("contents")

    output = str(output)
    flow_sampler.output = output
    expected_rf = str(expected_rf)
    assert os.path.exists(expected_rf)

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ) as mock_resume:
        out = FlowSampler._resume_from_file(
            flow_sampler,
            NestedSampler,
            resume_file=resume_file,
            model=integration_model,
            weights_path=None,
            flow_config=None,
        )
    mock_resume.assert_called_with(
        expected_rf,
        integration_model,
        weights_path=None,
        flow_config=None,
    )

    assert out is ns


@pytest.mark.integration_test
def test_init_cannot_resume_integration(tmp_path, integration_model):
    """Integration test for invalid resume files"""
    output = tmp_path / "test"
    output.mkdir()
    resume_file = "test.pkl"
    expected_rf = output / (resume_file + ".old")
    expected_rf.write_text("contents")
    side_effect = [RuntimeError, RuntimeError]
    output = str(output)
    expected_rf = str(expected_rf)

    assert os.path.exists(expected_rf)

    flow_sampler.output = output

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ), pytest.raises(
        RuntimeError, match=r"Could not resume sampler with error"
    ):
        FlowSampler(
            integration_model,
            output=output,
            resume_file=resume_file,
            resume_data=None,
            resume=True,
        )


def test_init_resume_error_no_file(flow_sampler, tmp_path):
    """Assert an error is raised if resume=True and a resume_file is \
        not specified.
    """
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)  #

    flow_sampler.check_resume = MagicMock(return_value=False)

    with patch("nessai.flowsampler.configure_threads"), patch(
        "nessai.flowsampler.NestedSampler"
    ) as mock_init:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            resume=True,
            resume_file=None,
        )

    mock_init.assert_called_once()


def test_resume_from_data(flow_sampler, model):
    """Test the resume from data method"""
    data = MagicMock()
    SamplerClass = MagicMock()
    weights_path = "test.pkl"
    flow_config = {"n_neurons": 24}
    FlowSampler._resume_from_data(
        flow_sampler,
        SamplerClass,
        data,
        model,
        weights_path,
        flow_config,
    )

    SamplerClass.resume_from_pickled_sampler.assert_called_once_with(
        data, model, weights_path=weights_path, flow_config=flow_config
    )


def test_init_signal_handling_enabled(flow_sampler, tmp_path):
    """Assert signal.signal is called when signal handling is enabled."""
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal") as mocked_fn:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            signal_handling=True,
        )
    mocked_fn.assert_called()


def test_init_signal_handling_disabled(flow_sampler, tmp_path):
    """Assert signal handling is not configure when disabled"""
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal") as mocked_fn:
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            signal_handling=False,
        )
    mocked_fn.assert_not_called()


def test_init_signal_handling_error(flow_sampler, tmp_path, caplog):
    """Assert signal handling is skipped if an error is raised."""
    integration_model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal", side_effect=AttributeError):
        FlowSampler.__init__(
            flow_sampler,
            integration_model,
            output=output,
            signal_handling=True,
        )
    assert "Cannot set signal attributes" in str(caplog.text)


def test_log_evidence(flow_sampler):
    """Test log-evidence property"""
    out = 0.0
    flow_sampler.logZ = out
    assert FlowSampler.log_evidence.__get__(flow_sampler) is out


def test_log_evidence_error(flow_sampler):
    """Test log-evidence property"""
    out = 0.1
    flow_sampler.logZ_error = out
    assert FlowSampler.log_evidence_error.__get__(flow_sampler) is out


@pytest.mark.parametrize("use_ins", [False, True])
def test_run(flow_sampler, use_ins):

    flow_sampler.importance_nested_sampler = use_ins

    kwargs = dict(
        plot=False,
        save=False,
        posterior_sampling_method="test",
        close_pool=True,
        plot_posterior=True,
    )
    FlowSampler.run(flow_sampler, **kwargs)

    if use_ins:
        flow_sampler.run_importance_nested_sampler.assert_called_once_with(
            **kwargs
        )
    else:
        flow_sampler.run_standard_sampler.assert_called_once_with(**kwargs)


@pytest.mark.parametrize("save", [False, True])
@pytest.mark.parametrize("plot", [False, True])
@patch(
    "nessai.flowsampler.draw_posterior_samples", return_value=np.array([0.1])
)
@patch("nessai.plot.plot_live_points")
@patch("nessai.plot.plot_indices")
def test_run_standard(
    mock_plot_indices, mock_plot_post, mock_draw_post, flow_sampler, save, plot
):
    """Test the run method"""
    nlive = 10
    log_Z = -5.0
    nested_samples = [0.1, 1.0, 10.0]
    log_w = np.array([-0.1, -0.2, -0.3])
    insertion_indices = [1, 2, 3]
    output = os.getcwd()
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.nlive = nlive
    flow_sampler.ns.insertion_indices = insertion_indices
    flow_sampler.output = output
    flow_sampler.ns.initialise = MagicMock()
    flow_sampler.ns.nested_sampling_loop = MagicMock(
        return_value=[log_Z, nested_samples]
    )
    flow_sampler.ns.state = MagicMock()
    flow_sampler.ns.state.log_posterior_weights = log_w
    flow_sampler.ns.state.plot_state = MagicMock()
    flow_sampler.save_results = MagicMock()
    flow_sampler.close_pool = True
    flow_sampler.result_extension = "hdf5"

    FlowSampler.run_standard_sampler(flow_sampler, save=save, plot=plot)

    flow_sampler.ns.initialise.assert_called_once()
    mock_draw_post.assert_called_once_with(
        nested_samples, log_w=log_w, method="rejection_sampling"
    )
    if save:
        flow_sampler.save_results.assert_called_once_with(
            os.path.join(output, "result"),
            extension="hdf5",
        )
    else:
        flow_sampler.save_results.assert_not_called()

    if plot:
        mock_plot_indices.assert_called_once_with(
            insertion_indices,
            nlive,
            filename=os.path.join(output, "insertion_indices.png"),
        )
        mock_plot_post.assert_called_once_with(
            flow_sampler.posterior_samples,
            filename=os.path.join(output, "posterior_distribution.png"),
        )
        flow_sampler.ns.state.plot.assert_called_once_with(
            os.path.join(output, "logXlogL.png"),
        )
    else:
        mock_plot_indices.assert_not_called()
        mock_plot_post.assert_not_called()
        flow_sampler.ns.state.plot.assert_not_called()

    assert flow_sampler.logZ == log_Z
    assert flow_sampler._nested_samples == nested_samples
    np.testing.assert_array_equal(
        flow_sampler.posterior_samples, np.array([0.1])
    )


@pytest.mark.parametrize("close_pool", [True, False, None])
def test_run_close_pool_standard(flow_sampler, close_pool):
    """Assert close pool is called following the intended logic"""
    flow_sampler.close_pool = True
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.close_pool = MagicMock()
    flow_sampler.ns.nested_sampling_loop = MagicMock(return_value=("lZ", "ns"))
    with patch("nessai.flowsampler.draw_posterior_samples"):
        FlowSampler.run_standard_sampler(
            flow_sampler, close_pool=close_pool, save=False, plot=False
        )

    # If True should be called
    # If None, should fallback to self.close_pool which is true
    if close_pool or close_pool is None:
        flow_sampler.ns.close_pool.assert_called_once()
    else:
        flow_sampler.ns.close_pool.assert_not_called()


@pytest.mark.parametrize("method", (None, "multinomial_resampling"))
def test_run_posterior_sampling_method_standard(flow_sampler, method):
    """Assert posterior sampling method is passed correctly"""
    log_w = np.random.rand(100)
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.nested_sampling_loop = MagicMock(return_value=("lZ", "ns"))
    flow_sampler.ns.state = MagicMock(spec=_NSIntegralState)
    flow_sampler.ns.state.log_posterior_weights = log_w
    with patch("nessai.flowsampler.draw_posterior_samples") as mock:
        FlowSampler.run_standard_sampler(
            flow_sampler,
            close_pool=False,
            save=False,
            plot=False,
            posterior_sampling_method=method,
        )

    if method is None:
        method = "rejection_sampling"
    mock.assert_called_once_with("ns", log_w=log_w, method=method)


def test_run_standard_plots_disabled(flow_sampler):
    """Assert individual plots can be disabled when plot=True"""
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.nested_sampling_loop = MagicMock(return_value=("lZ", "ns"))
    flow_sampler.ns.nlive = 100
    flow_sampler.state = MagicMock()
    flow_sampler.state.plot = MagicMock()

    with patch("nessai.flowsampler.draw_posterior_samples"), patch(
        "nessai.plot.plot_indices"
    ) as mock_indices, patch("nessai.plot.plot_live_points") as mock_post:
        FlowSampler.run_standard_sampler(
            flow_sampler,
            plot=True,
            save=False,
            close_pool=False,
            plot_indices=False,
            plot_posterior=False,
            plot_logXlogL=False,
        )
    mock_indices.assert_not_called()
    mock_post.assert_not_called()
    flow_sampler.state.plot.assert_not_called()


@pytest.mark.parametrize("close_pool", [False, True])
def test_run_ins(flow_sampler, close_pool):
    """Test running the importance nested sampler"""
    flow_sampler.importance_nested_sampler = True

    flow_sampler.close_pool = close_pool
    logZ = 0.0
    logZ_err = 0.1
    nested_samples = np.array([1, 2, 3])
    post = np.array([1, 2])
    ns = MagicMock()
    ns.nested_sampling_loop = MagicMock()
    ns.samples = nested_samples
    ns.log_evidence = logZ
    ns.log_evidence_error = logZ_err
    ns.draw_posterior_samples = MagicMock(return_value=post)
    ns.close_pool = MagicMock()
    flow_sampler.ns = ns

    FlowSampler.run_importance_nested_sampler(
        flow_sampler, redraw_samples=False, save=False, plot=False
    )

    ns.draw_posterior_samples.assert_called_once_with(
        sampling_method="importance_sampling",
        use_final_samples=False,
    )

    assert flow_sampler.logZ is logZ
    assert flow_sampler.logZ_error is logZ_err
    assert flow_sampler.initial_posterior_samples is post
    assert flow_sampler.posterior_samples is post
    assert flow_sampler._nested_samples is nested_samples
    if close_pool:
        ns.close_pool.assert_called_once()


def test_run_ins_redraw(flow_sampler):
    """Test running the importance nested sampler and redrawing samples"""
    flow_sampler.importance_nested_sampler = True
    flow_sampler.close_pool = False
    logZ = 0.0
    logZ_err = 0.1
    final_logZ = 0.01
    final_logZ_err = 0.11
    nested_samples = np.array([1, 2, 3])
    final_samples = np.array([4, 5, 6])
    post = np.array([1, 2])
    final_post = np.array([4, 5])
    ns = MagicMock()
    ns.nested_sampling_loop = MagicMock()
    ns.samples = nested_samples
    ns.log_evidence = logZ
    ns.log_evidence_error = logZ_err
    ns.final_log_evidence = final_logZ
    ns.final_log_evidence_error = final_logZ_err
    ns.draw_posterior_samples = MagicMock(
        side_effect=[post, final_post],
    )
    ns.draw_final_samples = MagicMock(return_value=(np.nan, final_samples))
    flow_sampler.ns = ns

    FlowSampler.run_importance_nested_sampler(
        flow_sampler,
        redraw_samples=True,
        save=False,
        plot=False,
        compute_initial_posterior=True,
    )

    ns.draw_posterior_samples.assert_has_calls(
        [
            call(
                sampling_method="importance_sampling", use_final_samples=False
            ),
            call(
                sampling_method="importance_sampling", use_final_samples=True
            ),
        ],
        any_order=False,
    )

    assert flow_sampler.initial_logZ is logZ
    assert flow_sampler.initial_logZ_error is logZ_err
    assert flow_sampler.initial_posterior_samples is post
    assert flow_sampler._nested_samples is nested_samples

    assert flow_sampler.logZ is final_logZ
    assert flow_sampler.logZ_error is final_logZ_err
    assert flow_sampler.posterior_samples is final_post


@pytest.mark.parametrize("test_class", [False, True])
def test_save_kwargs(flow_sampler, tmpdir, test_class):
    """Test the save kwargs method.

    If `test_class` is true, tests the case of the flow class being a class
    rather than a string.
    """
    kwargs = dict(nlive=10, a=np.array([0.1]))
    if test_class:
        from nessai.proposal import FlowProposal

        kwargs["flow_class"] = FlowProposal
    else:
        kwargs["flow_class"] = "flowproposal"

    flow_sampler.output = str(tmpdir.mkdir("test"))
    flow_sampler.eps = 1e-8
    flow_sampler.torch_dtype = "float32"
    flow_sampler.importance_nested_sampler = False

    FlowSampler.save_kwargs(flow_sampler, kwargs)

    assert os.path.exists(os.path.join(flow_sampler.output, "config.json"))


@pytest.mark.parametrize(
    "filename, extension", [("result", "json"), ("result.json", None)]
)
def test_save_result_json(
    flow_sampler, posterior_samples, filename, extension
):
    """Test saving with a JSON file."""
    d = dict(a=1)
    ns = MagicMock()
    ns.get_result_dictionary = MagicMock(return_value=d)
    flow_sampler.ns = ns
    flow_sampler.posterior_samples = posterior_samples

    with patch("nessai.flowsampler.save_to_json") as mock_save:
        FlowSampler.save_results(flow_sampler, filename, extension=extension)

    mock_save.assert_called_once_with(d, "result.json")


@pytest.mark.parametrize(
    "filename, extension", [("result", "hdf5"), ("result.hdf5", None)]
)
def test_save_result_hdf5(
    flow_sampler, posterior_samples, filename, extension
):
    """Test saving with an HDF5 file."""
    d = dict(a=1)
    ns = MagicMock()
    ns.get_result_dictionary = MagicMock(return_value=d)
    flow_sampler.ns = ns
    flow_sampler.posterior_samples = posterior_samples

    with patch("nessai.flowsampler.save_dict_to_hdf5") as mock_save:
        FlowSampler.save_results(flow_sampler, filename, extension=extension)

    mock_save.assert_called_once_with(d, "result.hdf5")


def test_save_result_no_extension(flow_sampler, posterior_samples):
    """Assert an error is raised if a file extension is not given or included
    in the filename.
    """
    d = dict(a=1)
    ns = MagicMock()
    ns.get_result_dictionary = MagicMock(return_value=d)
    flow_sampler.ns = ns
    flow_sampler.posterior_samples = posterior_samples
    with pytest.raises(
        RuntimeError,
        match=r"Must specify file extension if not present in filename!",
    ):
        FlowSampler.save_results(flow_sampler, "result")


def test_save_result_error(flow_sampler, posterior_samples):
    """Assert an error is raised if the extension is not recognised"""
    d = dict(a=1)
    ns = MagicMock()
    ns.get_result_dictionary = MagicMock(return_value=d)
    flow_sampler.ns = ns
    flow_sampler.posterior_samples = posterior_samples
    with pytest.raises(RuntimeError, match=r"Unknown file extension: pkl"):
        FlowSampler.save_results(flow_sampler, "result.pkl")


def test_terminate_run(flow_sampler):
    """Test terminate run"""
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.checkpoint = MagicMock()
    flow_sampler.ns.close_pool = MagicMock()
    FlowSampler.terminate_run(flow_sampler, code=2)
    flow_sampler.ns.checkpoint.assert_called_once()
    flow_sampler.ns.close_pool.assert_called_once_with(code=2)


def test_safe_exit(flow_sampler):
    """Test the safe exit method."""
    flow_sampler.exit_code = 130
    flow_sampler.terminate_run = MagicMock()

    with patch("sys.exit") as mock_exit:
        FlowSampler.safe_exit(flow_sampler, signum=2)

    mock_exit.assert_called_once_with(130)
    flow_sampler.terminate_run.assert_called_once_with(code=2)


@pytest.mark.parametrize("kwargs", [dict(n_pool=None), dict(n_pool=2)])
@pytest.mark.slow_integration_test
@pytest.mark.timeout(60)
@pytest.mark.skip_on_windows
def test_signal_handling(
    tmp_path, caplog, integration_model, kwargs, mp_context
):
    """Test the signal handling in nessai.

    Test is based on a similar test in bilby which is in turn based on: \
        https://stackoverflow.com/a/49615525/18400311
    """
    output = tmp_path / "output"
    output.mkdir()

    with patch("multiprocessing.Pool", mp_context.Pool), patch(
        "nessai.utils.multiprocessing.multiprocessing.get_start_method",
        mp_context.get_start_method,
    ):
        FlowSampler(
            integration_model,
            output=output,
            nlive=500,
            poolsize=1000,
            exit_code=5,
            signal_handling=True,
            **kwargs,
        )

    pid = os.getpid()

    def trigger_signal():
        time.sleep(2)
        os.kill(pid, signal.SIGINT)

    thread = Thread(target=trigger_signal)
    thread.daemon = True
    thread.start()

    with pytest.raises(SystemExit):
        try:
            time.sleep(4)
        except SystemExit as error:
            assert error.code == 5
            raise

    assert f"Trying to safely exit with code {signal.SIGINT}" in str(
        caplog.text
    )


@pytest.mark.slow_integration_test
@pytest.mark.timeout(60)
@pytest.mark.skip_on_windows
def test_signal_handling_disabled(tmp_path, caplog, integration_model):
    """Assert signal handling is correctly disabled.

    Test is based on a similar test in bilby which is in turn based on: \
        https://stackoverflow.com/a/49615525/18400311
    """
    output = tmp_path / "output"
    output.mkdir()

    FlowSampler(
        integration_model,
        output=output,
        nlive=500,
        poolsize=1000,
        exit_code=4,
        signal_handling=False,
    )

    pid = os.getpid()

    def trigger_signal():
        time.sleep(2)
        os.kill(pid, signal.SIGINT)

    thread = Thread(target=trigger_signal)
    thread.daemon = True
    thread.start()

    # Need to catch SIGINT to prevent tests for exiting
    def handler(signum=None, frame=None):
        sys.exit(10)

    signal.signal(signal.SIGINT, handler)

    with pytest.raises(SystemExit):
        try:
            time.sleep(4)
        except SystemExit as error:
            assert error.code == 10
            raise

    assert "Signal handling is disabled" in str(caplog.text)


@pytest.mark.parametrize("extension", ["json", "hdf5"])
@pytest.mark.integration_test
def test_save_results_integration(
    flow_sampler, tmpdir, posterior_samples, extension
):
    """Test the save results method"""
    output = str(tmpdir.mkdir("test"))
    filename = os.path.join(output, ".".join(["result", extension]))
    d = dict(a=1)

    ns = MagicMock()
    ns.get_result_dictionary = MagicMock(return_value=d)

    flow_sampler.ns = ns
    flow_sampler.posterior_samples = posterior_samples
    flow_sampler.initial_posterior_samples = posterior_samples.copy()

    FlowSampler.save_results(flow_sampler, filename)

    assert os.path.exists(filename)
    ns.get_result_dictionary.assert_called_once()

    if extension == "hdf5":
        out = h5py.File(filename, "r")
    else:
        with open(filename, "r") as fp:
            out = json.load(fp)

    assert "posterior_samples" in out
    np.testing.assert_array_equal(
        out["posterior_samples"]["x"], posterior_samples["x"]
    )


@pytest.mark.integration_test
@pytest.mark.parametrize("ins", [True, False])
def test_resume_from_data_integration(
    integration_model, tmp_path, caplog, ins
):
    output = tmp_path / "test"

    caplog.set_level(logging.INFO)

    kwargs = {}
    if ins:
        kwargs["min_samples"] = 4
        kwargs["max_iteration"] = 2
    else:
        kwargs["max_iteration"] = 10

    fs = FlowSampler(
        integration_model,
        nlive=10,
        output=output,
        checkpointing=False,
        resume_file=None,
        importance_nested_sampler=ins,
        **kwargs,
    )
    fs.run()

    pickled_sampler = pickle.dumps(fs.ns)
    resume_data = pickle.loads(pickled_sampler)

    fs_resume = FlowSampler(
        integration_model,
        nlive=10,
        output=output,
        checkpointing=False,
        resume_data=resume_data,
        resume_file=None,
        importance_nested_sampler=ins,
        **kwargs,
    )

    fs_resume.ns.iteration == fs.ns.iteration
    fs_resume.ns.log_evidence == fs.ns.log_evidence

    assert "Trying to resume sampler from `resume_data`" in str(caplog.text)



# -*- coding: utf-8 -*-
"""
Utilities related to histogramming.
"""
import numpy as np


def _hist_bin_fd(x):
    """
    The Freedman-Diaconis histogram bin estimator.

    See original Numpy implementation.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.
    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    iqr = np.subtract(*np.percentile(x, [75, 25]))
    return 2.0 * iqr * x.size ** (-1.0 / 3.0)


def _hist_bin_sturges(x):
    """
    Sturges histogram bin estimator.

    See original Numpy implementation.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.
    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    return np.ptp(x) / (np.log2(x.size) + 1.0)


def auto_bins(x, max_bins=50):
    """
    Compute the number bins for a histogram using numpy.histogram_bin_edges
    but enforces a maximum number of bins.

    Parameters
    ----------
    array : array_like
        Input data
    bins : int or sequence of scalars or str, optional
        Method for determining number of bins, see numpy documentation
    max_bins : int, optional (1000)
        Maximum number of bins

    Returns
    -------
    int
        Number of bins
    """
    x = np.asarray(x)
    if not x.size:
        raise RuntimeError("Input array is empty!")
    fd_bw = _hist_bin_fd(x)
    sturges_bw = _hist_bin_sturges(x)
    if fd_bw:
        bw = min(fd_bw, sturges_bw)
    else:
        bw = sturges_bw

    if bw:
        n_bins = int(np.ceil(np.ptp(x)) / bw)
    else:
        n_bins = 1

    nbins = min(n_bins, max_bins)
    return nbins



"""Tests for the config module"""

from nessai.config import LivepointsConfig
import numpy as np


def test_livepoint_config_reset_properties():
    """Assert the properties are reset"""
    conf = LivepointsConfig()

    assert conf.core_parameters == ["logP", "logL", "it"]
    assert conf.core_parameters_dtype == ["f8", "f8", "i4"]
    assert conf.core_parameters_defaults == (np.nan, np.nan, 0)
    assert conf.extra_parameters == []
    assert conf.extra_parameters_dtype == []
    assert conf.extra_parameters_defaults == ()

    assert conf.non_sampling_dtype == ["f8", "f8", "i4"]
    assert conf.non_sampling_defaults == (np.nan, np.nan, 0)

    conf.default_float_value = -np.inf
    conf.extra_parameters = ["a"]
    conf.extra_parameters_defaults = (0.0,)
    conf.extra_parameters_dtype = ["f4"]

    assert conf.non_sampling_defaults == (np.nan, np.nan, 0)

    conf.reset_properties()

    assert conf.core_parameters_defaults == (-np.inf, -np.inf, 0)
    assert conf.non_sampling_parameters == ["logP", "logL", "it", "a"]
    assert conf.non_sampling_dtype == ["f8", "f8", "i4", "f4"]
    assert conf.non_sampling_defaults == (-np.inf, -np.inf, 0, 0.0)


def test_livepoint_config_reset():
    """Assert the reset method clears the values"""
    conf = LivepointsConfig(
        extra_parameters=["a", "b"],
        extra_parameters_dtype=["f4", "i4"],
        extra_parameters_defaults=[0.0, 0],
    )
    assert conf.non_sampling_parameters == ["logP", "logL", "it", "a", "b"]
    conf.reset()
    assert conf.extra_parameters == []
    assert conf.extra_parameters_dtype == []
    assert conf.extra_parameters_defaults == ()
    assert conf.non_sampling_parameters == ["logP", "logL", "it"]



"""Test the properties defined in ImportanceFlowProposal"""

import pickle
from unittest.mock import MagicMock, patch

from nessai.proposal.importance import ImportanceFlowProposal as IFP
import pytest


def test_resume(ifp, model, tmp_path):

    flow_config = dict(patience=10, model_config=dict(n_inputs=2))

    ifp.initialise = MagicMock()
    ifp.flow = MagicMock()
    ifp.flow.resume = MagicMock()

    path = tmp_path / "test_resume"

    with patch("nessai.proposal.importance.Proposal.resume") as mock_parent:
        IFP.resume(ifp, model, flow_config, weights_path=path)

    mock_parent.assert_called_once_with(model)
    ifp.initialise.assert_called_once_with()

    assert ifp.flow_config is flow_config
    ifp.flow.resume.assert_called_once_with(
        flow_config["model_config"],
        weights_path=path,
    )


@pytest.mark.integration_test
@pytest.mark.usefixtures("ins_parameters")
def test_getstate_integration(tmp_path, model):

    ifp = IFP(
        model,
        output=tmp_path / "test_resume",
        weighted_kl=False,
    )
    ifp.initialise()
    weights = {-1: 1.0}

    for i in range(4):
        ifp.train(model.new_point(10), max_epochs=2)
        weights = {j - 1: 1 / (i + 2) for j in range(i + 2)}
        ifp.update_proposal_weights(weights)
        ifp.draw(10)

    out = pickle.dumps(ifp)

    ifp_resume = pickle.loads(out)
    ifp_resume.resume(model, {})

    assert len(ifp_resume.flow.models) == 4



# -*- coding: utf-8 -*-
"""
Integration tests for adding the default reparameterisations
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec

from nessai.model import Model
from nessai.proposal import FlowProposal
from nessai.reparameterisations import default_reparameterisations

# General reparameterisations that do not need extra parameters
general_reparameterisations = {
    k: v
    for k, v in default_reparameterisations.items()
    if k not in ["scale", "rescale", "angle-pair", "scaleandshift"]
}


@pytest.fixture(params=general_reparameterisations.keys())
def reparameterisation(request):
    return request.param


@pytest.fixture
def model():
    m = create_autospec(Model)
    m.names = ["x"]
    m.bounds = {"x": [-1, 1]}
    m.reparameterisations = None
    return m


@pytest.mark.integration_test
def test_configure_reparameterisations(tmpdir, model, reparameterisation):
    """Test adding one of the default reparameterisations.

    Only tests reparameterisations that don't need extra parameters.
    """
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir("test")),
        poolsize=10,
        reparameterisations={"x": reparameterisation},
    )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None


@pytest.mark.integration_test
@pytest.mark.parametrize("reparameterisation", ["scale", "rescale"])
def test_configure_reparameterisation_scale(tmpdir, model, reparameterisation):
    """Test adding the `Rescale` reparameterisation"""
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir("test")),
        poolsize=10,
        reparameterisations={
            "x": {"reparameterisation": reparameterisation, "scale": 2.0}
        },
    )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None


@pytest.mark.integration_test
def test_configure_reparameterisation_angle_pair(tmpdir, model):
    """Test adding the `AnglePair` reparameterisation"""
    model.names.append("y")
    model.bounds = {"x": [0, 2 * np.pi], "y": [-np.pi / 2, np.pi / 2]}
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir("test")),
        poolsize=10,
        reparameterisations={
            "x": {"reparameterisation": "angle-pair", "parameters": ["y"]}
        },
    )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None


@pytest.mark.integration_test
def test_default_reparameterisations(caplog, tmpdir):
    """Assert that by default the reparameterisation is RescaleToBounds"""
    caplog.set_level("INFO")
    model = MagicMock()
    model.names = ["x1", "x10", "x11"]
    model.bounds = {p: [-1, 1] for p in model.names}
    model.reparameterisations = None
    proposal = FlowProposal(
        model, poolsize=100, output=str(tmpdir.mkdir("test"))
    )
    # Mocked model so can't verify rescaling
    proposal.verify_rescaling = MagicMock()
    proposal.initialise()
    reparams = list(proposal._reparameterisation.values())
    assert len(reparams) == 1
    assert reparams[0].parameters == ["x1", "x10", "x11"]
    assert proposal.rescale_parameters == ["x1", "x10", "x11"]



"""Configuration for the reparameterisation tests"""

from dataclasses import dataclass
from typing import List

from nessai.livepoint import empty_structured_array
from nessai.utils.testing import assert_structured_arrays_equal
import numpy as np
import pytest


@dataclass
class _LightReparam:
    """Reparameterisation-like object"""

    name: str
    parameters: List[str]
    requires: List[str]


@pytest.fixture
def LightReparam() -> _LightReparam:
    return _LightReparam


@pytest.fixture()
def is_invertible(model, n=100):
    """Test if a reparameterisation is invertible."""

    def test_invertibility(reparam, model=model, atol=0.0, rtol=0.0):
        x = model.new_point(N=n)
        x_prime = empty_structured_array(n, names=reparam.prime_parameters)
        log_j = np.zeros(n)
        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

        x_in = empty_structured_array(x_re.size, reparam.parameters)
        log_j = np.zeros(x_re.size)

        x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
            x_in, x_prime_re, log_j
        )

        assert_structured_arrays_equal(x_inv, x, atol=atol, rtol=rtol)
        np.testing.assert_allclose(-log_j_inv, log_j_re)

        return True

    return test_invertibility



# -*- coding: utf-8 -*-
"""
Test the ScaleAndShift class.
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, call, create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.reparameterisations import ScaleAndShift
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture()
def reparam():
    return create_autospec(ScaleAndShift)


@pytest.fixture()
def parameters():
    return ["x", "y"]


@pytest.fixture()
def prior_bounds(parameters):
    return {p: [-1, 1] for p in parameters}


@pytest.mark.parametrize("scale", [2, 2.0, [1, 2], {"x": 1, "y": 2}])
def test_init_scale(scale, parameters, prior_bounds):
    """Test the init method with different input types"""
    reparam = ScaleAndShift(
        parameters=parameters, scale=scale, prior_bounds=prior_bounds
    )

    assert not set(reparam.scale.keys()) - set(parameters)
    assert isinstance(reparam.scale["x"], float)
    assert reparam.estimate_scale is False
    assert reparam.estimate_shift is False
    assert reparam._update is False
    assert reparam.shift is None


def test_init_scale_and_shift(reparam, parameters, prior_bounds):
    """Test the init with scale and shift"""
    scale = 1.0
    shift = 2.0
    scale_out = {"x": 1.0, "y": 1.0}
    shift_out = {"x": 2.0, "y": 2.0}
    reparam._check_value = MagicMock(side_effect=[scale_out, shift_out])
    ScaleAndShift.__init__(
        reparam,
        parameters=parameters,
        prior_bounds=prior_bounds,
        scale=scale,
        shift=shift,
    )

    reparam._check_value.assert_has_calls(
        [call(scale, "scale"), call(shift, "shift")]
    )
    assert reparam.scale == scale_out
    assert reparam.shift == shift_out
    assert reparam._update is False


def test_init_estimate(reparam, parameters, prior_bounds):
    """Assert estimate shift and scale are enabled"""
    ScaleAndShift.__init__(
        reparam,
        parameters=parameters,
        prior_bounds=prior_bounds,
        estimate_scale=True,
        estimate_shift=True,
    )
    assert reparam.estimate_scale is True
    assert reparam.estimate_shift is True
    assert list(reparam.scale) == parameters
    assert list(reparam.shift) == parameters
    assert all([v is None for v in reparam.scale.values()])
    assert all([v is None for v in reparam.shift.values()])
    assert reparam._update is True


@pytest.mark.parametrize("n", [1, 2])
def test_reparameterise_scale(reparam, n):
    """Test the reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    reparam.shift = None
    x = numpy_array_to_live_points(np.ones((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.zeros((n, 2)), reparam.prime_parameters
    )
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = ScaleAndShift.reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x, x_out)
    np.testing.assert_array_equal(log_j_out, -np.log(8 * np.ones(n)))
    assert (x_prime_out["x_prime"] == -0.5).all()
    assert (x_prime_out["y_prime"] == 0.25).all()


@pytest.mark.parametrize("n", [1, 2])
def test_reparameterise_scale_and_shift(reparam, n):
    """Test the reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    reparam.shift = {"x": 2.0, "y": -2.0}
    x = numpy_array_to_live_points(np.ones((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.zeros((n, 2)), reparam.prime_parameters
    )
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = ScaleAndShift.reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x, x_out)
    np.testing.assert_array_equal(log_j_out, -np.log(8 * np.ones(n)))
    assert (x_prime_out["x_prime"] == 0.5).all()
    assert (x_prime_out["y_prime"] == 0.75).all()


@pytest.mark.parametrize("scale", [1e60, 1e-60])
def test_reparameterise_scale_overflow(reparam, scale):
    """Test the reparameterise method with very small and large scales.

    Checks precision to 14 decimal places.
    """
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.scale = {"x": scale}
    reparam.shift = None
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(
        scale * x_array[:, np.newaxis], reparam.parameters
    )
    x_prime = numpy_array_to_live_points(
        np.ones((x_array.size, 1)), reparam.prime_parameters
    )
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = ScaleAndShift.reparameterise(
        reparam, x, x_prime, log_j
    )

    np.testing.assert_array_almost_equal(
        x_array, x_prime_out["x_prime"], decimal=14
    )
    assert (log_j == -np.log(scale)).all()


@pytest.mark.parametrize("n", [1, 2])
def test_inverse_reparameterise_scale(reparam, n):
    """Test the inverse reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    reparam.shift = None
    x = numpy_array_to_live_points(np.zeros((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.ones((n, 2)), reparam.prime_parameters
    )
    x_prime["x_prime"] *= -1
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = ScaleAndShift.inverse_reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x_prime, x_prime_out)
    np.testing.assert_array_equal(log_j_out, np.log(8 * np.ones(n)))
    assert (x_out["x"] == 2.0).all()
    assert (x_out["y"] == 4.0).all()


@pytest.mark.parametrize("scale", [1e60, 1e-60])
def test_inverse_reparameterise_scale_overflow(reparam, scale):
    """
    Test the inverse_reparameterise method with very small and large scales.
    """
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.scale = {"x": scale}
    reparam.shift = None
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(
        np.ones((x_array.size, 1)), reparam.parameters
    )
    x_prime = numpy_array_to_live_points(
        x_array[:, np.newaxis], reparam.prime_parameters
    )
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = ScaleAndShift.inverse_reparameterise(
        reparam, x, x_prime, log_j
    )

    np.testing.assert_array_equal(x_array * scale, x_out["x"])
    assert (log_j == np.log(scale)).all()


@pytest.mark.parametrize("n", [1, 2])
def test_inverse_reparameterise_scale_and_shift(reparam, n):
    """Test the inverse reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    reparam.shift = {"x": 1.0, "y": -2.0}
    x = numpy_array_to_live_points(np.zeros((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.ones((n, 2)), reparam.prime_parameters
    )
    x_prime["x_prime"] *= -1
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = ScaleAndShift.inverse_reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x_prime, x_prime_out)
    np.testing.assert_array_equal(log_j_out, np.log(8 * np.ones(n)))
    assert (x_out["x"] == 3.0).all()
    assert (x_out["y"] == 2.0).all()


@pytest.mark.parametrize("est_scale", [False, True])
@pytest.mark.parametrize("est_shift", [False, True])
def test_update_both(reparam, parameters, est_scale, est_shift):
    """Assert update updates the correct values"""
    x = np.random.randn(4, 2).astype(dtype=[(p, "f8") for p in parameters])
    reparam.parameters = parameters
    reparam.scale = {p: None for p in parameters}
    reparam.shift = {p: None for p in parameters}
    reparam._update = True
    reparam.estimate_scale = est_scale
    reparam.estimate_shift = est_shift
    with patch("numpy.std", side_effect=[1, 2]), patch(
        "numpy.mean", side_effect=[3, 4]
    ):
        ScaleAndShift.update(reparam, x)

    if est_scale:
        assert reparam.scale == dict(zip(parameters, [1, 2]))
    else:
        assert all([v is None for v in reparam.scale.values()])

    if est_shift:
        assert reparam.shift == dict(zip(parameters, [3, 4]))
    else:
        assert all([v is None for v in reparam.shift.values()])


def test_init_no_scale():
    """Make sure an error is raised if the scale is not given"""
    with pytest.raises(
        RuntimeError, match="Must specify a scale or enable estimate_scale"
    ):
        ScaleAndShift(scale=None, estimate_scale=False)


@pytest.mark.parametrize("scale", [[1], [1, 2, 3]])
def test_init_incorrect_scale_list(scale):
    """Make sure an error is raised if the scale is the incorrect length"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        ScaleAndShift(
            parameters=parameters, scale=scale, prior_bounds=prior_bounds
        )

    assert "different length" in str(excinfo.value)


@pytest.mark.parametrize("scale", [{"x": 1}, {"x": 1, "y": 1, "z": 1}])
def test_init_incorrect_scale_dict(scale):
    """Make sure an error is raised if the scale keys to not match the \
            parameters.
    """
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        ScaleAndShift(
            parameters=parameters, scale=scale, prior_bounds=prior_bounds
        )

    assert "Mismatched parameters" in str(excinfo.value)


def test_init_incorrect_scale_type():
    """Make sure an error is raised if the scale is the incorrect type"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(TypeError, match=r"scale input must be .*"):
        ScaleAndShift(
            parameters=parameters, scale="1", prior_bounds=prior_bounds
        )


def test_init_incorrect_shift_type():
    """Make sure an error is raised if the scale is the incorrect type"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(TypeError, match=r"shift input must be .*"):
        ScaleAndShift(
            parameters=parameters,
            scale=1,
            shift="1",
            prior_bounds=prior_bounds,
        )


@pytest.mark.integration_test
@pytest.mark.parametrize("scale", [-2.0, 2.0])
@pytest.mark.parametrize("shift", [-2.0, 2.0, None])
@pytest.mark.parametrize("estimate_scale", [True, False])
@pytest.mark.parametrize("estimate_shift", [True, False])
def test_invertible(
    is_invertible, model, scale, shift, estimate_scale, estimate_shift
):
    """Assert scale and shift is invertible for different settings."""

    x = model.new_point(100)

    reparam = ScaleAndShift(
        parameters=model.names,
        prior_bounds=model.bounds,
        scale=scale,
        shift=shift,
        estimate_scale=estimate_scale,
        estimate_shift=estimate_shift,
    )

    reparam.update(x)
    assert is_invertible(reparam, atol=1e-14)



# -*- coding: utf-8 -*-
"""Tests for the optimisation utilities."""
from collections import namedtuple
from unittest.mock import patch

import numpy as np
import pytest

from nessai.livepoint import numpy_array_to_live_points
from nessai.utils.testing import assert_structured_arrays_equal
from nessai.utils.optimise import optimise_meta_proposal_weights

n_its = 10


@pytest.mark.usefixtures("ins_parameters")
@pytest.mark.parametrize("initial_weights", [None, np.arange(n_its)])
def test_optimise_meta_weights(initial_weights):
    """Assert correct method from scipy is called"""
    n = 100
    samples = numpy_array_to_live_points(np.random.randn(n, 2), ["x0", "x1"])
    samples["logL"] = np.random.randn(n)
    samples["logQ"] = np.random.randn(n)
    samples["logW"] = -samples["logW"]
    samples["it"] = np.random.randint(-1, n_its, size=n)
    log_q = np.random.randn(n, n_its + 1)

    input_samples = samples.copy()

    expected = np.random.rand(n_its + 1)

    Result = namedtuple("Result", ["x"])
    result = Result(x=expected)

    def check_loss(loss, *args, **kwargs):
        value = loss(np.random.rand(n_its + 1))
        assert np.isfinite(value)
        return result

    with patch(
        "nessai.utils.optimise.minimize", side_effect=check_loss
    ) as mock:
        out = optimise_meta_proposal_weights(
            samples, log_q=log_q, initial_weights=initial_weights
        )

    mock.assert_called_once()
    np.testing.assert_array_equal(out, result.x)

    # Make sure inputs are unchanged
    assert_structured_arrays_equal(samples, input_samples)



# -*- coding: utf-8 -*-
"""
Test utilities for sampling in the latent space.
"""
import numpy as np
import pytest
from scipy import stats, special
from unittest.mock import create_autospec, patch

from nessai.utils.sampling import (
    NDimensionalTruncatedGaussian,
    compute_radius,
    draw_gaussian,
    draw_nsphere,
    draw_surface_nsphere,
    draw_truncated_gaussian,
    draw_uniform,
)


def test_compute_radius():
    """Assert compute radius calls the correct function"""
    with patch("scipy.stats.chi.ppf") as mock:
        compute_radius(2, 0.95)
    mock.assert_called_once_with(0.95, 2)


@pytest.mark.parametrize(
    "d, q, r",
    [
        [1, 0.6827, 1.0],
        [1, 0.9545, 2.0],
        [2, 0.3935, 1.0],
        [5, 0.8909, 3.0],
        [10, 0.9004, 4.0],
    ],
)
def test_compute_radius_integration(d, q, r):
    """Assert compute radius returns the expected value.

    Checks that for a given confidence level the correct radius is returned,
    values checked correspond to standard deviations.

    Values for q taken from: \
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4358977/
    """
    r_out = compute_radius(d, q)
    np.testing.assert_almost_equal(r_out, r, decimal=4)


@pytest.mark.parametrize("ndims, radius", [(2, 1), (10, 2), (10, 10), (1, 1)])
def test_draw_surface_nsphere(ndims, radius):
    """
    Assert the correct number of samples with the correct dimensions are drawn
    and all lie on the surface.
    """
    out = draw_surface_nsphere(ndims, r=radius, N=1000)

    assert out.shape[0] == 1000
    assert out.shape[1] == ndims
    np.testing.assert_allclose(np.sqrt(np.sum(out**2.0, axis=1)), radius)


@pytest.mark.parametrize("ndims, radius", [(2, 1), (10, 2), (10, 10), (1, 1)])
def test_draw_nball(ndims, radius):
    """
    Assert the correct number of samples with the correct dimensions are drawn
    and all lie within the n-ball.
    """
    out = draw_nsphere(ndims, r=radius, N=1000)

    assert out.shape[0] == 1000
    assert out.shape[1] == ndims
    np.testing.assert_array_less(np.sqrt(np.sum(out**2, axis=-1)), radius)


def test_draw_uniform():
    """Assert the underlying numpy function is called correctly."""
    expected = np.array([0.5, 1])
    with patch("numpy.random.uniform", return_value=expected) as mock:
        out = draw_uniform(2, r=1, N=100, fuzz=2)
    mock.assert_called_once_with(0, 1, (100, 2))
    np.testing.assert_array_equal(out, expected)


def test_draw_gaussian():
    """Assert the underlying numpy function is called correctly."""
    expected = np.array([1, 2])
    with patch("numpy.random.randn", return_value=expected) as mock:
        out = draw_gaussian(2, r=1, N=100, fuzz=2)
    mock.assert_called_once_with(100, 2)
    np.testing.assert_array_equal(out, expected)


@pytest.mark.parametrize(
    "r, var, fuzz",
    [
        (1.0, 1.0, 1.0),
        (2.0, 1.0, 1.0),
        (2.0, 2.0, 1.0),
        (4.0, 2.0, 1.5),
        (7.0, 4.0, 2.0),
    ],
)
@pytest.mark.flaky(reruns=5)
def test_draw_truncated_gaussian_1d(r, var, fuzz):
    """
    Test drawing from a truncated Gaussian in 1d
    """
    s = draw_truncated_gaussian(1, r, var=var, N=2000, fuzz=fuzz)
    sigma = np.sqrt(var)
    d = stats.truncnorm(
        -r * fuzz / sigma, r * fuzz / sigma, loc=0, scale=sigma
    )
    _, p = stats.kstest(np.squeeze(s), d.cdf)
    assert p >= 0.05


@pytest.mark.parametrize("dims", [1, 2, 4, 8, 16, 32, 64])
@pytest.mark.parametrize("radius", [1.0, 2.0, 4.0])
@pytest.mark.parametrize("fuzz", [1.0, 1.1, 1.5])
def test_ndimensional_truncated_gaussian_u_max(dims, radius, fuzz):
    """Check the value of the u_max"""
    dist = create_autospec(NDimensionalTruncatedGaussian)

    expected_u_max = special.gammainc(dims / 2, (radius * fuzz) ** 2 / 2)

    NDimensionalTruncatedGaussian.__init__(dist, dims, radius, fuzz)

    assert dist.u_max == expected_u_max


@pytest.mark.parametrize(
    "r, fuzz",
    [
        (1.0, 1.0),
        (2.0, 1.0),
        (2.0, 1.0),
        (4.0, 1.5),
        (7.0, 2.0),
    ],
)
@pytest.mark.flaky(reruns=5)
def test_ndimensional_truncated_gaussian_sample(r, fuzz):
    """Assert samples are correctly drawn from the distribution"""
    dist = NDimensionalTruncatedGaussian(1, r, fuzz=fuzz)
    s = dist.sample(10_000)
    d = stats.truncnorm(-r * fuzz, r * fuzz, loc=0)
    _, p = stats.kstest(np.squeeze(s), d.cdf)
    assert p >= 0.05



# -*- coding: utf-8 -*-
"""
Test the functions related to yielding new samples
"""
import pytest
from unittest.mock import MagicMock

from nessai.livepoint import parameters_to_live_point
from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture
def old_sample(model):
    s = parameters_to_live_point([5, 5], names=model.names)
    s["logL"] = 0.0
    s["logP"] = 0.0
    return s


@pytest.fixture
def sampler(sampler, old_sample):
    sampler.proposal = MagicMock()
    sampler.logLmin = sampler.model.log_likelihood(old_sample)
    sampler.logLmax = sampler.logLmin
    return sampler


def test_yield_sample_accept(sampler, old_sample):
    """Test function when sample is accepted"""
    new_sample = parameters_to_live_point([1, 1], names=sampler.model.names)
    new_sample["logL"] = 0.0
    new_sample["logP"] = 0.0
    sampler.proposal.draw = MagicMock(return_value=new_sample)

    count, next_sample = next(NestedSampler.yield_sample(sampler, old_sample))

    assert count == 1
    sampler.proposal.draw.assert_called_once_with(old_sample)
    assert next_sample == new_sample
    assert sampler.logLmax == sampler.model.log_likelihood(new_sample)


def test_yield_sample_reject(sampler, old_sample):
    """Test function when sample is rejected and a new sample is drawn"""
    new_samples = [
        parameters_to_live_point([6, 6], names=sampler.model.names),
        parameters_to_live_point([1, 1], names=sampler.model.names),
    ]
    for s in new_samples:
        s["logL"] = 0.0
        s["logP"] = 0.0

    sampler.proposal.draw = MagicMock(side_effect=new_samples)
    sampler.proposal.populated = True

    count, next_sample = next(NestedSampler.yield_sample(sampler, old_sample))

    assert count == 2
    assert next_sample == new_samples[1]
    sampler.proposal.draw.assert_called_with(old_sample)


def test_yield_sample_not_populated(sampler, old_sample):
    """Test function when sample is rejected and a new sample is not drawn"""
    new_sample = parameters_to_live_point([6, 6], names=sampler.model.names)

    sampler.proposal.draw = MagicMock(return_value=new_sample)
    sampler.proposal.populated = False

    count, next_sample = next(NestedSampler.yield_sample(sampler, old_sample))

    assert count == 1
    assert old_sample == next_sample
    sampler.proposal.draw.assert_called_once_with(old_sample)



# -*- coding: utf-8 -*-
"""
Functions and objects related to reparametersiations for use in the
``reparameterisations`` dictionary.

See the documentation for an in-depth description of how to use these
functions and classes.
"""
from .angle import Angle, AnglePair, ToCartesian
from .base import Reparameterisation
from .combined import CombinedReparameterisation
from .null import NullReparameterisation
from .rescale import Rescale, RescaleToBounds, ScaleAndShift
from .utils import get_reparameterisation


default_reparameterisations = {
    "default": (RescaleToBounds, None),
    "rescaletobounds": (RescaleToBounds, None),
    "rescale-to-bounds": (RescaleToBounds, None),
    "offset": (RescaleToBounds, {"offset": True}),
    "inversion": (
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "split",
        },
    ),
    "inversion-duplicate": (
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "duplicate",
        },
    ),
    "logit": (
        RescaleToBounds,
        {
            "rescale_bounds": [0.0, 1.0],
            "update_bounds": False,
            "post_rescaling": "logit",
        },
    ),
    "log-rescale": (
        RescaleToBounds,
        {
            "rescale_bounds": [0.0, 1.0],
            "update_bounds": False,
            "post_rescaling": "log",
        },
    ),
    "scale": (Rescale, None),
    "scaleandshift": (ScaleAndShift, None),
    "rescale": (Rescale, None),
    "zscore": (
        ScaleAndShift,
        {"estimate_scale": True, "estimate_shift": True},
    ),
    "z-score": (
        ScaleAndShift,
        {"estimate_scale": True, "estimate_shift": True},
    ),
    "angle": (Angle, {}),
    "angle-pi": (Angle, {"scale": 2.0, "prior": "uniform"}),
    "angle-2pi": (Angle, {"scale": 1.0, "prior": "uniform"}),
    "angle-sine": (RescaleToBounds, None),
    "angle-cosine": (RescaleToBounds, None),
    "angle-pair": (AnglePair, None),
    "periodic": (Angle, {"scale": None}),
    "to-cartesian": (ToCartesian, None),
    "none": (NullReparameterisation, None),
    "null": (NullReparameterisation, None),
    None: (NullReparameterisation, None),
}


__all__ = [
    "Angle",
    "AnglePair",
    "CombinedReparameterisation",
    "NullReparameterisation",
    "Reparameterisation",
    "Rescale",
    "RescaleToBounds",
    "ToCartesian",
    "get_reparameterisation",
]



#!/usr/bin/env python

# Example of using nessai

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

# Setup the logger - credit to the Bilby team for this neat function!
# see: https://git.ligo.org/lscsoft/bilby

output = "./outdir/2d_gaussian_example/"
logger = setup_logger(output=output)

# Define the model, in this case we use a simple 2D gaussian
# The model must contain names for each of the parameters and their bounds
# as a dictionary with arrays/lists with the min and max

# The main functions in the model should be the log_prior and log_likelihood
# The log prior must be able to accept structured arrays of points
# where each field is one of the names in the model and there are two
# extra fields which are `logP` and `logL'


class GaussianModel(Model):
    """A simple two-dimensional Gaussian likelihood."""

    def __init__(self):
        # Names of parameters to sample
        self.names = ["x", "y"]
        # Prior bounds for each parameter
        self.bounds = {"x": [-10, 10], "y": [-10, 10]}

    def log_prior(self, x):
        """
        Returns log of prior given a live point assuming uniform
        priors on each parameter.
        """
        # Check if values are in bounds, returns True/False
        # Then take the log to get 0/-inf and make sure the dtype is float
        log_p = np.log(self.in_bounds(x), dtype="float")
        # Iterate through each parameter (x and y)
        # since the live points are a structured array we can
        # get each value using just the name
        for n in self.names:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        """
        Returns log likelihood of given live point assuming a Gaussian
        likelihood.
        """
        log_l = np.zeros(x.size)
        # Use a Gaussian logpdf and iterate through the parameters
        for n in self.names:
            log_l += norm.logpdf(x[n])
        return log_l


# The FlowSampler object is used to managed the sampling. Keyword arguments
# are passed to the nested sampling.
fs = FlowSampler(GaussianModel(), output=output, resume=False, seed=1234)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""Different nested samplers available in nessai.

All samplers inherit from :py:obj:`nessai.samplers.base.BaseNestedSampler`.
"""
from .nestedsampler import NestedSampler
from .importancesampler import ImportanceNestedSampler

__all__ = [
    "ImportanceNestedSampler",
    "NestedSampler",
]



# -*- coding: utf-8 -*-
"""
Utilities related to logging.
"""
import logging
import os
import sys


def setup_logger(
    output=None,
    label="nessai",
    log_level="INFO",
    filehandler_kwargs=None,
    stream=None,
):
    """
    Setup the logger.

    Based on the implementation in Bilby:
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils/log.py

    Parameters
    ----------
    output : str, optional
        Path of to output directory.
    label : str, optional
        Label for this instance of the logger.
    log_level : {'ERROR', 'WARNING', 'INFO', 'DEBUG'}, optional
        Level of logging passed to logger.
    filehandler_kwargs : dict, optional
        Keyword arguments for configuring the FileHandler. See logging
        documentation for details.
    stream : str, file-object, optional
        Stream passes to :code:`logging.StreamHandler` to set the stream. See
        logging documentation for more details.

    Returns
    -------
    :obj:`logging.Logger`
        Instance of the Logger class.
    """
    from .. import __version__ as version

    if isinstance(log_level, str):
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError("log_level {} not understood".format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger("nessai")
    logger.setLevel(level)

    if (
        any([isinstance(h, logging.StreamHandler) for h in logger.handlers])
        is False
    ):
        if isinstance(stream, str):
            if stream.lower() == "stderr":
                stream = sys.stderr
            elif stream.lower() == "stdout":
                stream = sys.stdout
            else:
                raise ValueError(
                    f"Unknown stream: {stream}. Choose from: [stderr, stdout]"
                )
        stream_handler = logging.StreamHandler(stream)
        stream_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(name)s %(levelname)-8s: %(message)s",
                datefmt="%m-%d %H:%M",
            )
        )
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if (
        any([isinstance(h, logging.FileHandler) for h in logger.handlers])
        is False
    ):
        if label:
            if output:
                if not os.path.exists(output):
                    os.makedirs(output, exist_ok=True)
            else:
                output = "."
            log_file = os.path.join(output, f"{label}.log")
            if filehandler_kwargs is None:
                filehandler_kwargs = {}
            file_handler = logging.FileHandler(log_file, **filehandler_kwargs)
            file_handler.setFormatter(
                logging.Formatter(
                    "%(asctime)s %(levelname)-8s: %(message)s", datefmt="%H:%M"
                )
            )

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    logger.info(f"Running Nessai version {version}")

    return logger



# -*- coding: utf-8 -*-
"""
Test plotting in the nested sampler.
"""
import os
import numpy as np
import pytest
from unittest.mock import MagicMock, patch
from nessai.samplers.nestedsampler import NestedSampler


@pytest.mark.parametrize("track_gradients", [False, True])
@pytest.mark.parametrize("filename", [None, "test.png"])
def test_plot_state(sampler, tmpdir, filename, track_gradients):
    """Test making the state plot"""
    x = np.arange(10)
    sampler.iteration = 1003
    sampler.train_on_empty = False
    sampler.history = dict(
        iterations=np.arange(10),
        min_log_likelihood=x,
        max_log_likelihood=x,
        logZ=x,
        dlogZ=x,
        likelihood_evaluations=x,
        sampling_time=x,
        training_iterations=[256, 711],
        population_iterations=[256, 500, 711, 800],
        population_acceptance=4 * [0.5],
        population_radii=4 * [1.0],
        checkpoint_iterations=[600],
        mean_acceptance=1 / x,
        rolling_p=np.arange(4),
    )
    sampler.state = MagicMock()
    sampler.state.log_vols = np.linspace(0, -10, 1050)
    sampler.state.track_gradients = track_gradients
    sampler.state.gradients = np.arange(1050)

    if filename is not None:
        sampler.output = tmpdir.mkdir("test_plot_state")
        filename = os.path.join(sampler.output, filename)
    fig = NestedSampler.plot_state(sampler, filename)

    if filename is not None:
        assert os.path.exists(filename)
    else:
        assert fig is not None


@pytest.mark.parametrize("samples", [[], [1, 2, 3]])
@pytest.mark.parametrize("filename", [None, "trace.png"])
@pytest.mark.parametrize("live_points", [True, False])
@pytest.mark.parametrize("trace_parameters", [None, ["x", "y"]])
@patch("nessai.samplers.nestedsampler.plot_trace", return_value="fig")
def test_plot_trace(
    mock_plot,
    sampler,
    tmpdir,
    samples,
    live_points,
    trace_parameters,
    filename,
):
    """Test the plot_trace method"""
    sampler.nested_samples = samples
    sampler.state = MagicMock
    sampler.state.log_vols = [1, 2, 3, 4]
    sampler.output = os.getcwd()
    sampler.nlive = 10
    logx_live = list(5 + np.arange(sampler.nlive))
    sampler.state.get_logx_live_points = MagicMock(return_value=logx_live)
    sampler.trace_parameters = trace_parameters

    if live_points:
        sampler.live_points = list(np.random.randn(sampler.nlive))
    else:
        sampler.live_points = None

    if filename is not None:
        sampler.output = tmpdir.mkdir("test_plot_trace")
        filename = os.path.join(sampler.output, filename)

    fig = NestedSampler.plot_trace(sampler, filename=filename)

    if not len(samples):
        mock_plot.assert_not_called()
        assert fig is None
    else:
        if live_points:
            sampler.state.get_logx_live_points.assert_called_once_with(
                sampler.nlive
            )
        else:
            sampler.state.get_logx_live_points.assert_not_called()
        mock_plot.assert_called_once_with(
            [2, 3, 4],
            samples,
            parameters=trace_parameters,
            live_points=sampler.live_points,
            log_x_live_points=logx_live if live_points else None,
            filename=filename,
        )
        assert fig == "fig"


@pytest.mark.parametrize("filename", [None, "trace.png"])
@patch("nessai.samplers.nestedsampler.plot_indices", return_value="fig")
def test_plot_insertion_indices(mock_plot, sampler, filename):
    """Test plotting the insertion indices"""
    nlive = 10
    indices = list(range(20))
    sampler.nlive = nlive
    sampler.insertion_indices = indices
    fig = NestedSampler.plot_insertion_indices(
        sampler, filename=filename, k=True
    )
    assert fig == "fig"
    mock_plot.assert_called_once_with(
        indices, nlive, filename=filename, k=True
    )



"""Test the INS evidence"""

from unittest.mock import MagicMock, create_autospec

import numpy as np
import pytest
from scipy.special import logsumexp

from nessai.evidence import (
    _INSIntegralState as INSState,
    log_evidence_from_ins_samples,
)


@pytest.fixture
def state():
    return create_autospec(INSState)


@pytest.fixture
def live_points(model, ins_parameters):
    x = model.new_point(2)
    x["logP"] = model.log_prior(x)
    x["logL"] = model.log_likelihood(x)
    x["logQ"] = model.new_point_log_prob(x)
    x["logW"] = x["logP"] - x["logQ"]
    return x


@pytest.fixture
def nested_samples(model, ins_parameters):
    x = model.new_point(4)
    x["logP"] = model.log_prior(x)
    x["logL"] = model.log_likelihood(x)
    x["logQ"] = model.new_point_log_prob(x)
    x["logW"] = x["logP"] - x["logQ"]
    return x


def test_init(state):
    """Test the init method"""
    INSState.__init__(state)
    assert state._logZ == np.NINF


def test_log_Z(state):
    """Assert the normalisation is applied"""
    state._logZ = -10.0
    state._n = 100
    assert INSState.logZ.__get__(state) == (-10.0 - np.log(100))


def test_log_evidence_error(state):
    """Assert property calls the the correct method"""
    state.compute_uncertainty = MagicMock(return_value=0.1)
    assert INSState.log_evidence_error.__get__(state) == 0.1
    state.compute_uncertainty.assert_called_once_with(log_evidence=True)


def test_evidence(state):
    state.log_evidence = 0.0
    assert INSState.evidence.__get__(state) == 1.0


def test_evidence_error(state):
    out = 0.1
    state.compute_uncertainty = MagicMock(return_value=out)
    assert INSState.evidence_error.__get__(state) == out
    state.compute_uncertainty.assert_called_once_with(log_evidence=False)


def test_log_evidence_live_points(state):
    """Check the correct value is returned"""
    state._weights_lp = np.log(np.array([1, 2, 3]))
    np.testing.assert_almost_equal(
        INSState.log_evidence_live_points.__get__(state), np.log(2)
    )


def test_log_evidence_live_points_error(state):
    """Assert an error is raised if the live points are None"""
    state._weights_lp = None
    with pytest.raises(RuntimeError, match="Live points are not set"):
        INSState.log_evidence_live_points.__get__(state)


def test_log_evidence_live_nested_samples(state):
    """Check the correct value is returned"""
    state._weights_ns = np.log(np.array([1, 2, 3]))
    np.testing.assert_almost_equal(
        INSState.log_evidence_nested_samples.__get__(state), np.log(2)
    )


def test_log_posterior_weights(state):
    """Check the posterior weights are correct"""
    state._weights = np.array([1, 2, 3])
    state.logZ = -1.0
    out = INSState.log_posterior_weights.__get__(state)
    np.testing.assert_equal(out, np.array([2, 3, 4]))


def test_update_evidence_ns_obly(state, nested_samples):
    """Test updating the evidence"""
    x = nested_samples
    expected = logsumexp(x["logL"] + x["logW"])
    assert np.isfinite(expected)
    INSState.update_evidence(state, nested_samples, live_points=None)
    np.testing.assert_equal(state._logZ, expected)
    assert state._n == x.size
    assert state._weights_lp is None


def test_update_evidence_both(state, nested_samples, live_points):
    """Test updating the evidence"""
    x = np.concatenate([nested_samples, live_points])
    expected = logsumexp(x["logL"] + x["logW"])
    print(x)
    assert np.isfinite(expected)
    INSState.update_evidence(state, nested_samples, live_points=live_points)
    np.testing.assert_equal(state._logZ, expected)
    assert state._n == x.size
    assert state._weights_lp is not None


@pytest.mark.parametrize("ns_only, expected", [(True, 1.0), (False, 2.0)])
def test_compute_evidence_ratio(state, ns_only, expected):
    """Assert the correct value is returned"""
    state.log_evidence_live_points = -1.0
    state.log_evidence_nested_samples = -2.0
    state.logZ = -3.0
    assert INSState.compute_evidence_ratio(state, ns_only) == expected


def test_compute_uncertainty_log(state):
    """Assert a finite value is returned"""
    state._weights = -np.array([1, 2, 3])
    state._n = 3
    state.logZ = -6.0
    out_ln = INSState.compute_uncertainty(state, log_evidence=True)
    out = INSState.compute_uncertainty(state, log_evidence=False)
    # Check errors are equivalent
    np.testing.assert_almost_equal(out_ln, out / np.exp(state.logZ))


def test_log_evidence_ins_samples():
    n = 10
    log_l = np.log(np.random.rand(n))
    log_w = np.log(np.random.rand(n))
    samples = np.array(
        [*zip(log_l, log_w)], dtype=[("logL", "f8"), ("logW", "f8")]
    )

    expected = np.log(np.mean(np.exp(log_l + log_w)))

    out = log_evidence_from_ins_samples(samples)

    np.testing.assert_almost_equal(out, expected, decimal=12)



# -*- coding: utf-8 -*-
"""
Distributions to use as the 'base distribution' for normalising flows.
"""
import math

from glasflow.distributions import (
    ResampledGaussian as BaseResampledGaussian,
)
from glasflow.nflows.distributions import Distribution
from glasflow.nflows.utils import torchutils
import numpy as np
import torch


class MultivariateNormal(Distribution):
    """
    A multivariate Normal with zero mean and specified covariance.

    Parameters
    ----------
    shape : tuple
        Shape of distribution, this is used to determine the number of
        dimensions.
    var : float, optional
        Variance of the distribution.
    """

    def __init__(self, shape, var=1):
        super().__init__()
        self._shape = torch.Size(shape)
        self._var = var
        self._std = math.sqrt(var)

        self.register_buffer(
            "_log_z",
            torch.tensor(
                0.5 * np.prod(shape) * np.log(2 * np.pi * var),
                dtype=torch.float64,
            ),
            persistent=False,
        )

    def _log_prob(self, inputs, context):
        # Note: the context is ignored.
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )
        neg_energy = -(0.5 / self._var) * torchutils.sum_except_batch(
            inputs**2, num_batch_dims=1
        )
        return neg_energy - self._log_z

    def _sample(self, num_samples, context):
        if context is None:
            return torch.normal(
                0,
                self._std,
                size=(num_samples, *self._shape),
                device=self._log_z.device,
            )
        else:
            raise NotImplementedError

    def _mean(self, context):
        if context is None:
            return self._log_z.new_zeros(self._shape)
        else:
            raise NotImplementedError


class ResampledGaussian(BaseResampledGaussian):
    """Wrapper for ResampledGaussian.

    Adds methods needed in nessai.
    """

    end_iteration = BaseResampledGaussian.estimate_normalisation_constant
    """Function to be called at the end of an iteration.

    For LARS this updates the estimate of the normalisation constant
    independently of the other parameters in the flow.
    """

    def finalise(self, n_samples: int = 10_000, n_batches: int = 10) -> None:
        """Finalise the estimate of the normalisation constant."""
        self.estimate_normalisation_constant(
            n_samples=n_samples, n_batches=n_batches
        )
